"""Collection of utilities of the freva command line argument parser."""

from __future__ import annotations
import argparse
from configparser import ConfigParser, ExtendedInterpolation
from copy import copy
from getpass import getuser
import logging
from pathlib import Path
from typing import Callable, Optional

import freva
from evaluation_system.misc import logger, config

subparser_func_type = Callable[
    [str, argparse._SubParsersAction], Optional["BaseParser"]
]
"""Type for a method that creates a sub-command parser. This method gets a string
representing the description of the sub command as well as the SubParserAction
this sub command parser is added to.
"""


def is_admin(raise_error: bool = False) -> bool:
    """Check if the user at runtime is one of the admins.

    Parameters:
    -----------
    raise_error:
        Raise a RuntimeError if user is not admin
    """
    config.reloadConfiguration()
    admin = [a for a in config.get("admins", "").split(",") if a.strip()]
    user = getuser()
    is_admin = user in admin
    if not is_admin and raise_error:
        raise RuntimeError(f"{user} is not in admin list")
    return is_admin


class BaseParser:
    """Base class for common command line argument parsers."""

    parser_func = argparse.ArgumentParser.parse_args
    """Define the standard arparse parsing function"""

    def __init__(
        self,
        sub_commands: dict[str, subparser_func_type],
        parser: argparse.ArgumentParser,
    ) -> None:
        """Create the sub-command parsers."""

        self.parser = parser
        self.subparsers = parser.add_subparsers(help="Available sub-commands:")
        self.help = self.get_subcommand_help()
        self.sub_commands = sub_commands
        for cmd, subparser in self.sub_commands.items():
            subparser(self.help[cmd], self.subparsers)

    @property
    def logger(self) -> logging.Logger:
        """Use evaluation_system logger in all classes using ths class."""
        return logger

    def set_debug(self, debug: bool) -> None:
        """Set the logger level to DEBUG."""
        if debug is True:
            self.logger.setLevel(logging.DEBUG)

    def parse_args(self, argv: Optional[list[str]] = None) -> argparse.Namespace:
        """Parse the command line arguments."""
        args, other_args = self.parser.parse_known_args(argv)
        self.kwargs = {k: v for (k, v) in args._get_kwargs() if k != "apply_func"}
        self.kwargs["other_args"] = other_args
        self.set_debug(self.kwargs.pop("debug", False))
        return args

    @staticmethod
    def parse_crawl_my_data(help: str, subparsers: argparse._SubParsersAction) -> None:
        """Parse the user data crawl."""
        from .crawl_my_data import CrawlDataCli

        call_parsers = subparsers.add_parser(
            "crawl-my-data",
            description=help,
            help=help,
            formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        )
        CrawlDataCli("freva", call_parsers)

    @staticmethod
    def parse_history(help: str, subparsers: argparse._SubParsersAction) -> None:
        """Parse the history command."""
        from .history import HistoryCli

        call_parser = subparsers.add_parser(
            "history",
            description=help,
            help=help,
            formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        )
        HistoryCli("freva", call_parser)

    @staticmethod
    def parse_plugin(help, subparsers: argparse._SubParsersAction) -> None:
        """Parse the plugin command."""
        from .plugin import PluginCli

        call_parser = subparsers.add_parser(
            "plugin",
            description=help,
            help=help,
            formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        )
        PluginCli("freva", call_parser)

    @staticmethod
    def parse_check(help: str, subparsers: argparse._SubParsersAction) -> None:
        """Parse the check command."""
        from .admin.check import CheckCli

        call_parser = subparsers.add_parser(
            "check",
            description=help,
            help=help,
            formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        )
        CheckCli(call_parser)

    @staticmethod
    def parse_solr(help: str, subparsers: argparse._SubParsersAction) -> None:
        """Parse the solr index command."""
        from .admin.solr import SolrCli

        call_parser = subparsers.add_parser(
            "solr",
            description=help,
            help=help,
            formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        )
        SolrCli(call_parser)

    @staticmethod
    def parse_doc(help: str, subparser: argparse._SubParsersAction) -> None:
        """Parse the docu update command."""
        from .admin.doc import DocCli

        call_parser = subparser.add_parser(
            "doc",
            description=help,
            help=help,
            formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        )
        DocCli(call_parser)

    @staticmethod
    def parse_esgf(help: str, subparsers: argparse._SubParsersAction) -> None:
        """Parse the esgf command."""
        from .esgf import EsgfCli

        call_parsers = subparsers.add_parser(
            "esgf",
            description=help,
            help=help,
            formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        )
        EsgfCli("freva", call_parsers)

    @staticmethod
    def parse_databrowser(help: str, subparsers: argparse._SubParsersAction) -> None:
        """Parse the databrowser command."""
        from .databrowser import DataBrowserCli

        call_parsers = subparsers.add_parser(
            "databrowser",
            description=help,
            help=help,
            formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        )
        DataBrowserCli("freva", call_parsers)

    @classmethod
    def get_subcommand_parsers(cls) -> dict[str, subparser_func_type]:
        """Create the help strings of the available sub commands."""
        sub_commands: dict[str, subparser_func_type] = {
            "databrowser": cls.parse_databrowser,
            "plugin": cls.parse_plugin,
            "history": cls.parse_history,
            "crawl-my-data": cls.parse_crawl_my_data,
            "esgf": cls.parse_esgf,
        }
        admin_commands: dict[str, subparser_func_type] = {
            "solr": cls.parse_solr,
            "check": cls.parse_check,
            "doc": cls.parse_doc,
        }
        if is_admin():
            return {**sub_commands, **admin_commands}
        return sub_commands

    @classmethod
    def get_subcommand_help(cls) -> dict[str, str]:
        """Create the help strings of the available sub commands."""
        sub_commands = {
            "databrowser": "Find data in the system.",
            "plugin": "Apply data analysis plugin.",
            "history": "Read the plugin application history.",
            "crawl-my-data": "Update users project data",
            "esgf": "Search/Download ESGF the data catalogue.",
        }
        admin_commands = {
            "solr": "Apache solr server related sub-commands.",
            "check": "Perform various checks.",
            "doc": "Update the plugin documentation.",
            "index": "(Re)-Index data on the apache solr server.",
            "broken-runs": "Check for broken runs and report them.",
            "pull-request": "Check for incoming pull requests.",
        }
        if is_admin():
            return {**sub_commands, **admin_commands}
        return sub_commands

    def _usage(self, *args: Optional[str], **kwargs: Optional[str]) -> None:
        """Exit with usage message."""
        self.parser.error(
            "the following sub-commands are "
            f"required: {', '.join(self.sub_commands.keys())}"
        )


class BaseCompleter:
    """Base class for command line argument completers."""

    def __init__(
        self,
        metavar: str,
        argv: list[str],
        choices: Optional[dict[str, tuple[str, str]]] = None,
        shell: str = "bash",
        strip: bool = False,
        flags_only: bool = False,
    ):
        self.choices = choices or {}
        self.strip = strip
        self.metavar = metavar
        self.argv = argv
        self.flags_only = flags_only
        if shell == "zsh":
            self.get_print = self._print_zsh
        elif shell == "fish":
            self.get_print = self._print_fish
        else:
            self.get_print = self._print_default

    def _print_zsh(self, choices: dict[str, tuple[str, str]]) -> list[str]:
        out = []
        for key, (help, func) in choices.items():
            if self.metavar != "databrowser" or key.startswith("-"):
                out.append(f"{key}[{help}]{func}")
            else:
                out.append(f"{key}: {help}")
        return out

    def _print_fish(self, choices: dict[str, tuple[str, str]]) -> list[str]:
        out = []
        for key, (help, func) in choices.items():
            out.append(f"{key}: {help}")
        return out

    def _print_default(self, choices: dict[str, tuple[str, str]]) -> list[str]:

        out = []
        for key, (help, func) in choices.items():
            if self.metavar == "databrowser" and not key.startswith("-"):
                out.append(f"{key}: {help}")
            else:
                out.append(key)
        return out

    def _get_databrowser_choices(self) -> dict[str, tuple[str, str]]:
        """Get the choices for databrowser command."""
        from freva import databrowser

        facet_args = []
        for arg in self.argv:
            try:
                key, value = arg.split("=")
            except ValueError:
                continue
            facet_args.append(arg)
        facets = BaseCompleter.arg_to_dict(facet_args)
        search = databrowser(attributes=False, all_facets=True, **facets)
        choices = {}
        for att, values in search.items():
            if att not in facets:
                keys = ",".join([v for n, v in enumerate(values)])
                choices[att] = (keys, "")
        return choices

    def _get_plugin_choices(self) -> dict[str, tuple[str, str]]:
        """Get the choices for the plugin command."""

        docs: dict[str, dict[str, str]] = {}
        desc: dict[str, dict[str, str]] = {}
        plugins: dict[str, str] = {}

        for plugin in freva.read_plugin_cache():
            plugins[plugin.name] = plugin.description
            docs[plugin.name] = {}
            desc[plugin.name] = {}
            for param, doc, value in plugin.parameters:
                desc[plugin.name][param] = doc
                docs[plugin.name][param] = value
        args = [arg for arg in self.argv if not arg.startswith("-") and arg != "plugin"]
        choices = {}
        if not args:
            # No plugin name was given
            return {pl: (help, "") for pl, help in plugins.items()}
        try:
            config = docs[args[0]]
            setup = desc[args[0]]
        except KeyError:
            # Wrong plugin name
            return {}
        options = []
        for key in config.keys():
            option_present = False
            for arg in args:
                if arg.startswith(f"{key}="):
                    option_present = True
                    break
            if not option_present:
                options.append(f"{key}=")
        for option in options:
            opt = option.strip("=")
            if "file" in setup[opt].lower() or "dir" in setup[opt].lower():
                choices[opt] = (setup[opt], ":file:_files")
            else:
                choices[opt] = (setup[opt], "")
        return choices

    @property
    def command_choices(self) -> dict[str, tuple[str, str]]:

        choices = {}
        if self.flags_only:
            return self.choices
        if self.metavar == "databrowser":
            choices = self._get_databrowser_choices()
        elif self.metavar == "plugin":
            choices = self._get_plugin_choices()
        return {**self.choices, **choices}

    def formated_print(self) -> None:
        """Print all choices to be processed by the shell completion function."""

        out = self.get_print(self.command_choices)
        for line in out:
            if line.startswith("-") and self.strip and not self.flags_only:
                continue
            elif not line.startswith("-") and self.flags_only:
                continue
            print(line)

    @staticmethod
    def arg_to_dict(
        args: Optional[list[str]], append: bool = False
    ) -> dict[str, list[str]]:
        """Convert a parsed arguments with key=value pairs to a dictionary.

        Parameters:
        ----------
        args:
            Collection of arguments that are converted to dict, the arguments
            should be of form key=value
        append:
            If a key occurs twice convert first value to list and append second
            value to list, the default behaviour is updating the entry with
            the second value
        Returns:
        --------
        dict: Dictionariy representation of key=value pairs
        """
        out_dict: dict[str, list[str]] = {}
        for arg in args or []:
            key, _, value = arg.partition("=")
            if append and key in out_dict:
                out_dict[key].append(value)
            else:
                out_dict[key] = [value]
        return out_dict

    @staticmethod
    def _get_choices_from_parser(
        parser: argparse.ArgumentParser,
    ) -> dict[str, tuple[str, str]]:

        choices = {}
        for action in parser._actions:
            if action.help == argparse.SUPPRESS:
                # This is an option that is not exposed to users
                continue
            action_type = ""
            if action.type == Path:
                action_type = ":file:_files"
            if action.option_strings:
                choice = action.option_strings[0]
            else:
                choice = action.dest
            if choice not in ("facet", "facets", "tool-name", "options"):
                choices[choice] = (action.help or "", action_type)
        return choices

    @classmethod
    def get_args_of_subcommand(cls, argv: list[str]) -> dict[str, tuple[str, str]]:
        """Get all possible arguments from a freva sub-command."""
        from freva.cli import crawl_my_data, databrowser, history, plugin, esgf

        sub_command = argv.pop(0)
        modules = {
            "plugin": plugin,
            "databrowser": databrowser,
            "crawl-my-data": crawl_my_data,
            "history": history,
            "esgf": esgf,
        }
        try:
            mod = modules[sub_command]
        except AttributeError:
            return {}
        CliParser = getattr(mod, mod.CLI)()
        return cls._get_choices_from_parser(CliParser.parser)

    @classmethod
    def parse_choices(cls, argv: list[str]) -> BaseCompleter:
        """Create the completion choices from given cmd arguments."""
        parser = argparse.ArgumentParser(
            description="Get choices for command line arguments"
        )
        parser.add_argument(
            "command", help="First command, freva, freva-histor etc", type=str
        )
        parser.add_argument("--shell", help="Shell in use", default="bash", type=str)
        parser.add_argument(
            "--strip",
            help="Do not print options starting with -",
            default=False,
            action="store_true",
        )
        parser.add_argument(
            "--flags-only",
            help="Only print options starting with -",
            default=False,
            action="store_true",
        )
        ap, args = parser.parse_known_args(argv)
        main_choices = {k: (v, "") for k, v in BaseParser.get_subcommand_help().items()}
        if ap.command == "freva" and not args:
            return cls(ap.command, [], main_choices, shell=ap.shell, strip=ap.strip)
        elif ap.command != "freva":
            sub_cmd = "-".join(ap.command.split("-")[1:])
            args.insert(0, sub_cmd)
        choices = {}
        if args[0] in main_choices:
            choices = cls.get_args_of_subcommand(copy(args))
        choices = {k: v for (k, v) in choices.items() if k not in args}
        return cls(
            args[0],
            args,
            choices,
            shell=ap.shell,
            strip=ap.strip,
            flags_only=ap.flags_only,
        )


def print_choices(arguments: list[str]) -> None:
    """Print completion choices based of command line arguments.

    Parameters:
    -----------
    arguments:
        List of command line arguments, that are evaluated by the choice printer
    """
    argv = [arg.strip() for arg in arguments if arg.strip()]
    comp = BaseCompleter.parse_choices(argv)
    comp.formated_print()



import os
from shutil import rmtree
import sys
from textwrap import dedent
import tempfile

import pytest

from evaluation_system.api.workload_manager.lsf import LSFJob, lsf_format_bytes_ceil
from evaluation_system.api.workload_manager.core import parse_bytes


def test_header():
    with LSFJob(walltime="00:02", processes=4, cores=8, memory="8GB") as cluster:

        assert "#BSUB" in cluster.job_header
        assert "#BSUB -J worker" in cluster.job_header
        assert "#BSUB -n 8" in cluster.job_header
        assert "#BSUB -M 8000" in cluster.job_header
        assert "#BSUB -W 00:02" in cluster.job_header
        assert "#BSUB -q" not in cluster.job_header
        assert "#BSUB -P" not in cluster.job_header

    with LSFJob(
        queue="general",
        project="DaskOnLSF",
        processes=4,
        cores=8,
        memory="28GB",
        ncpus=24,
        mem=100000000000,
    ) as cluster:

        assert "#BSUB -q general" in cluster.job_header
        assert "#BSUB -J worker" in cluster.job_header
        assert "#BSUB -n 24" in cluster.job_header
        assert "#BSUB -n 8" not in cluster.job_header
        assert "#BSUB -M 100000" in cluster.job_header
        assert "#BSUB -M 28000" not in cluster.job_header
        assert "#BSUB -W" in cluster.job_header
        assert '#BSUB -P "DaskOnLSF"' in cluster.job_header

    with LSFJob(
        queue="general",
        project="Dask On LSF",
        processes=4,
        cores=8,
        memory="28GB",
        ncpus=24,
        mem=100000000000,
    ) as cluster:

        assert "#BSUB -q general" in cluster.job_header
        assert "#BSUB -J worker" in cluster.job_header
        assert "#BSUB -n 24" in cluster.job_header
        assert "#BSUB -n 8" not in cluster.job_header
        assert "#BSUB -M 100000" in cluster.job_header
        assert "#BSUB -M 28000" not in cluster.job_header
        assert "#BSUB -W" in cluster.job_header
        assert '#BSUB -P "Dask On LSF"' in cluster.job_header

    with LSFJob(cores=4, memory="8GB") as cluster:

        assert "#BSUB -n" in cluster.job_header
        assert "#BSUB -W" in cluster.job_header
        assert "#BSUB -M" in cluster.job_header
        assert "#BSUB -q" not in cluster.job_header
        assert "#BSUB -P" not in cluster.job_header

    with LSFJob(cores=4, memory="8GB", job_extra=["-u email@domain.com"]) as cluster:

        assert "#BSUB -u email@domain.com" in cluster.job_header
        assert "#BSUB -n" in cluster.job_header
        assert "#BSUB -W" in cluster.job_header
        assert "#BSUB -M" in cluster.job_header
        assert "#BSUB -q" not in cluster.job_header
        assert "#BSUB -P" not in cluster.job_header


def test_job_script():
    with LSFJob(walltime="00:02", processes=4, cores=8, memory="28GB") as cluster:

        job_script = cluster.job_script()
        assert "#BSUB" in job_script
        assert "#BSUB -J worker" in job_script
        assert "#BSUB -n 8" in job_script
        assert "#BSUB -M 28000" in job_script
        assert "#BSUB -W 00:02" in job_script
        assert "#BSUB -q" not in cluster.job_header
        assert "#BSUB -P" not in cluster.job_header

    with LSFJob(
        queue="general",
        project="DaskOnLSF",
        processes=4,
        cores=8,
        memory="28GB",
        ncpus=24,
        mem=100000000000,
    ) as cluster:

        job_script = cluster.job_script()
        assert "#BSUB -q general" in cluster.job_header
        assert "#BSUB -J worker" in cluster.job_header
        assert "#BSUB -n 24" in cluster.job_header
        assert "#BSUB -n 8" not in cluster.job_header
        assert "#BSUB -M 100000" in cluster.job_header
        assert "#BSUB -M 28000" not in cluster.job_header
        assert "#BSUB -W" in cluster.job_header
        assert '#BSUB -P "DaskOnLSF"' in cluster.job_header

    with LSFJob(
        walltime="1:00",
        cores=1,
        memory="16GB",
        project="Dask On LSF",
        job_extra=["-R rusage[mem=16GB]"],
    ) as cluster:

        job_script = cluster.job_script()

        assert "#BSUB -J worker" in cluster.job_header
        assert "#BSUB -n 1" in cluster.job_header
        assert "#BSUB -R rusage[mem=16GB]" in cluster.job_header
        assert "#BSUB -M 16000000" in cluster.job_header
        assert "#BSUB -W 1:00" in cluster.job_header
        assert '#BSUB -P "Dask On LSF"' in cluster.job_header


def test_informative_errors():
    with pytest.raises(ValueError) as info:
        LSFJob(memory=None, cores=4)
    assert "memory" in str(info.value)


def lsf_unit_detection_helper(expected_unit, conf_text=None):
    temp_dir = tempfile.mkdtemp()
    current_lsf_envdir = os.environ.get("LSF_ENVDIR", None)
    os.environ["LSF_ENVDIR"] = temp_dir
    if conf_text is not None:
        with open(os.path.join(temp_dir, "lsf.conf"), "w") as conf_file:
            conf_file.write(conf_text)
    memory_string = "13GB"
    memory_base = parse_bytes(memory_string)
    correct_memory = lsf_format_bytes_ceil(memory_base, lsf_units=expected_unit)
    with LSFJob(memory=memory_string, cores=1) as cluster:
        assert "#BSUB -M %s" % correct_memory in cluster.job_header
    rmtree(temp_dir)
    if current_lsf_envdir is None:
        del os.environ["LSF_ENVDIR"]
    else:
        os.environ["LSF_ENVDIR"] = current_lsf_envdir


@pytest.mark.parametrize(
    "lsf_units_string,expected_unit",
    [
        ("LSF_UNIT_FOR_LIMITS=MB", "mb"),
        ("LSF_UNIT_FOR_LIMITS=G  # And a comment", "gb"),
        ("#LSF_UNIT_FOR_LIMITS=NotDetected", "kb"),
    ],
)
def test_lsf_unit_detection(lsf_units_string, expected_unit):
    conf_text = dedent(
        """
        LSB_JOB_MEMLIMIT=Y
        LSB_MOD_ALL_JOBS=N
        LSF_PIM_SLEEPTIME_UPDATE=Y
        LSF_PIM_LINUX_ENHANCE=Y
        %s
        LSB_DISABLE_LIMLOCK_EXCL=Y
        LSB_SUBK_SHOW_EXEC_HOST=Y
        """
        % lsf_units_string
    )
    lsf_unit_detection_helper(expected_unit, conf_text)


def test_lsf_unit_detection_without_file():
    lsf_unit_detection_helper("kb", conf_text=None)



from django.db import models
from django.contrib.auth.models import User
from evaluation_system.model.plugins.models import Version, Parameter
from evaluation_system.misc.utils import PrintableList
import re
import os
import json


class History(models.Model):
    """
    The class belongs to a table containing all processes, which were started with analyze.
    """

    class Meta:
        """
        Set the user's permissions
        """

        permissions = (
            ("history_submit_job", "Can submit a job"),
            ("history_cancel_job", "Can cancel a job"),
            ("browse_full_data", "Can search all data"),
        )
        app_label = "history"

    class processStatus:
        """
        The allowed statuses
        finished           - the process finished and produced output files
        finished_no_output - the process finished, but no output files were created
        scheduled          - the job was send to slurm
        running            - the job is executed
        broken             - an exception occurred
        not_scheduled      - an error occurred during scheduling
        """

        finished, finished_no_output, broken, running, scheduled, not_scheduled = list(
            range(6)
        )

    class Flag:
        """
        The possible flags are:
        free    - can be accessed by all
        guest   - like public and permission for guest users
        public  - the data is accessible for every registered user
        shared  - the data is can be only accessed by certain users (to be implemented)
        private - the data is private
        deleted - the data set will be hidden
        """

        public, shared, private, deleted = range(4)
        guest = 8
        free = 9

    STATUS_CHOICES = (
        (processStatus.finished, "finished"),
        (processStatus.finished_no_output, "finished (no output)"),
        (processStatus.broken, "broken"),
        (processStatus.running, "running"),
        (processStatus.scheduled, "scheduled"),
        (processStatus.not_scheduled, "not scheduled"),
    )

    FLAG_CHOICES = (
        (Flag.public, "public"),
        (Flag.shared, "shared"),
        (Flag.private, "private"),
        (Flag.deleted, "deleted"),
        (Flag.guest, "users and guest"),
        (Flag.free, "no login required"),
    )

    #: Date and time when the process were scheduled
    timestamp = models.DateTimeField()
    #: Name of the tool
    tool = models.CharField(max_length=50)
    #: Version of the tool
    version = models.CharField(max_length=20)
    #: User ID
    version_details = models.ForeignKey(Version, on_delete=models.CASCADE, default=1)
    #: The configuration this can be quiet lengthy
    configuration = models.TextField()
    #: Output file generated by SLURM
    slurm_output = models.TextField()
    #: IP addr of the host submitting a job
    host = models.TextField()
    #: User ID
    uid = models.ForeignKey(
        User, on_delete=models.CASCADE, to_field="username", db_column="uid"
    )
    #: Status (scheduled, running, finished, cancelled)
    status = models.IntegerField(choices=STATUS_CHOICES)
    #: Flag (deleted, private, shared, public)
    flag = models.IntegerField(choices=FLAG_CHOICES, default=Flag.public)
    # User defined caption for the analysis
    caption = models.CharField(max_length=255, blank=True, null=False)

    def __init__(self, *args, **kwargs):
        """
        Creates a dictionary for projectStatus
        """
        self.status_dict = dict()
        public_props = (
            name for name in dir(self.processStatus) if not name.startswith("_")
        )
        for name in public_props:
            self.status_dict[getattr(self.processStatus, name)] = name

        super(History, self).__init__(*args, **kwargs)

    def __str__(self, compact=True):  # pragma: no cover
        conf_str = ""

        if compact:
            conf_str = str(self.configuration)
            if len(conf_str) > 70:
                conf_str = conf_str[:67] + "..."
            version = ""
        else:
            items = ["%15s=%s" % (k, v) for k, v in sorted(self.config_dict().items())]
            if items:
                # conf_str = '\n' + json.dumps(self.configuration, sort_keys=True, indent=2)
                conf_str = "\nConfiguration:\n%s" % "\n".join(items)
            # if self.results:
            #     conf_str = '%s\nOutput:\n%s' % (conf_str, '\n'.join(out_files))

            version = "%s %s" % (
                self.version,
                self.version_details.internal_version_tool,
            )

        return "%s) %s%s [%s] %s %s" % (
            self.pk,
            self.tool,
            version,
            self.timestamp,
            self.status_name(),
            conf_str,
        )

    def slurmId(self):
        id = re.sub(r".*\-", "", self.slurm_output)
        id = re.sub(r"\..*", "", id)

        # always return a number, even when the string is too short
        # (the default value for the string is '0')
        if not id:
            id = "0"

        return id

    def config_dict(self, load_default_values=False):
        """
        Converts the configuration to a dictionary
        """

        d = {}

        config = Configuration.objects.filter(history_id_id=self.id).order_by("pk")

        for c in config:
            name = c.parameter_id.parameter_name
            if load_default_values and c.is_default:
                d[name] = json.loads(c.parameter_id.default)
            else:
                d[name] = json.loads(c.value)

            # we have to take care that some special values are casted to the right class
            if c.parameter_id.parameter_type == "Range" and d[name] is not None:
                d[name] = PrintableList(d[name])

        return d

    def status_name(self):
        """
        Returns status as string
        """
        return self.status_dict[self.status]

    def get_slurm_status(self):
        """
        Method to get the slurm status.
        """
        slurm_id = self.slurmId()
        # not started with slurm
        if slurm_id == "0":
            return False
        try:
            cmd = "sacct -X -p -j %s" % slurm_id
            p = os.popen(cmd, "r")
            header = p.readline().split("|")
            entry = p.readline().split("|")
            index = header.index("State")
            return entry[index]
        # If "sacct" is not configured, we have to use "squeue"
        except ValueError:
            cmd = "squeue -h"
            p = os.popen(cmd, "r")
            while True:
                line = p.readline()
                if not line:
                    if self.status == 3:  # self.STATUS_CHOICES.running:
                        return "Cancelled"
                    return self.status_name()
                if slurm_id in line:
                    if " PD " in line:
                        return "Scheduled"
                    return "Running"

    @staticmethod
    def find_similar_entries(
        config, uid=None, max_impact=Parameter.Impact.affects_plots, max_entries=-1
    ):
        """
        Find entries which are similar to a given configuration.
        :param config: The configuration as array.
        :type config: array of history_configuration objects
        :param uid: the users id to find private results
        :type uid: str
        :param max_impact: The maximal impact level recognized
        :type max_impact: integer
        :param max_entries: The maximal number of results to be returned
        :type max_entries: integer
        """

        from django.db.models import Count, Q

        o = Configuration.objects.all()

        length = 0

        parameter = None

        # We use django Q to create the query.
        # this routine builds the parameter to query.
        for c in config:
            if c.parameter_id.impact <= max_impact:
                # both parameter and value have to match
                andparam = Q(parameter_id_id=c.parameter_id) & Q(value=c.value)

                # concatenate all parameter pairs with an or condition
                if parameter is None:
                    parameter = andparam
                else:
                    parameter = parameter | andparam

                length += 1

        if parameter is not None:
            o = o.filter(parameter)
            o = o.values("history_id_id").annotate(hcount=Count("history_id"))

            # using a less than equal relation would allow to access matches
            # which are equal to n percent.
            o = o.filter(hcount=length).order_by("-history_id_id")
        else:
            o = []

        # there should be an easier method to get a list the ids of the found
        # datasets
        history_list = []

        valid_status = [
            History.processStatus.finished,
            History.processStatus.finished_no_output,
        ]

        public_flags = [History.Flag.free, History.Flag.public]

        private_flags = [
            History.Flag.private,
            History.Flag.deleted,
            History.Flag.shared,
        ]

        for row in o:
            h = History.objects.filter(pk=row["history_id_id"])[0]

            if h.status in valid_status and (
                h.flag in public_flags or (h.uid_id == uid and h.flag in private_flags)
            ):
                # append entry to list and stop if the desired list length is reached
                history_list.append(h)

                if len(history_list) == max_entries:
                    break

        return history_list


class Result(models.Model):
    """
    This class belongs to a table storing results.
    The output files of process will be recorded here.
    """

    class Filetype:
        """
        Different IDs of file types
        data      - ascii or binary data to download
        plot      - a file which can be converted to a picture
        preview   - a local preview picture (copied or converted)
        """

        data, plot, preview = range(3)
        unknown = 9

    FILE_TYPE_CHOICES = (
        (Filetype.data, "data"),
        (Filetype.plot, "plot"),
        (Filetype.preview, "preview"),
        (Filetype.unknown, "unknown"),
    )

    #: history id
    history_id = models.ForeignKey(History, on_delete=models.CASCADE)
    #: path to the output file
    output_file = models.TextField()
    #: path to preview file
    preview_file = models.TextField(default="")
    #: specification of a file type
    file_type = models.IntegerField(choices=FILE_TYPE_CHOICES)

    class Meta:
        """
        Set the user's permissions
        """

        permissions = (("results_view_others", "Can view results from other users"),)
        app_label = "history"

    def fileExtension(self):
        """
        Returns the file extension of the result file
        """
        from os import path

        return path.splitext(self.output_file)[1]

    # some not yet implemented ideas
    #: Allows a logical clustering of results
    # group           = models.IntegerField(max_length=2)
    #: Defines an order for each group
    # group_order     = models.IntegerField(max_length=2)


class ResultTag(models.Model):
    """
    This class belongs to a table storing results.
    The output files of process will be recorded here.
    """

    class flagType:
        caption = 0

    class Meta:
        app_label = "history"

    TYPE_CHOICES = ((flagType.caption, "Caption"),)

    #: result id
    result_id = models.ForeignKey(Result, on_delete=models.CASCADE)
    #: specification of a file type
    type = models.IntegerField(choices=TYPE_CHOICES)
    #: path to the output file
    text = models.TextField()


class HistoryTag(models.Model):
    """
    This class belongs to a table storing results.
    The output files of process will be recorded here.
    """

    class tagType:
        [caption, note_public, note_private, note_deleted, follow, unfollow] = list(
            range(6)
        )

    class Meta:
        app_label = "history"

    TYPE_CHOICES = (
        (tagType.caption, "Caption"),
        (tagType.note_public, "Public note"),
        (tagType.note_private, "Private note"),
        (tagType.note_deleted, "Deleted note"),
        (tagType.follow, "Follow"),
        (tagType.unfollow, "Unfollow"),
    )

    #: result id
    history_id = models.ForeignKey(History, on_delete=models.CASCADE)
    #: specification of a file type
    type = models.IntegerField(choices=TYPE_CHOICES)
    #: path to the output file
    text = models.TextField()
    #: the user, who tagged the history entry
    uid = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
        to_field="username",
        db_column="uid",
        null=True,
        default=None,
    )


class Configuration(models.Model):
    """
    Holds the configuration
    """

    #: history id
    history_id = models.ForeignKey(
        History, on_delete=models.CASCADE, related_name="history_id"
    )

    #: parameter number
    parameter_id = models.ForeignKey(
        Parameter, on_delete=models.CASCADE, related_name="parameter_id"
    )

    #: md5 checksum of value (not used, yet)
    md5 = models.CharField(max_length=32, default="")

    #: value
    value = models.TextField(null=True, blank=True)

    # is the default value used?
    is_default = models.BooleanField()

    class Meta:
        app_label = "history"



"""Provide different utilities that does not depend on any other internal package."""
from __future__ import annotations
import copy
from copy import deepcopy
from difflib import get_close_matches
import errno
import os
from re import split
import shlex
from subprocess import run, PIPE
from string import Template
from typing import Any, Dict, Iterable, List, TextIO, IO, Union


def run_cmd(cmd: str, **kwargs: Any) -> str:
    """Run a command via subprocess.run"""
    kwargs.setdefault("check", False)
    kwargs["stdout"] = kwargs["stderr"] = PIPE
    res = run(shlex.split(cmd), **kwargs)
    return res.stdout.decode()


def get_console_size() -> Dict[str, int]:
    """Try getting the size of the current tty."""
    console_size = run_cmd("stty size")
    rows, columns = 25, 80
    try:
        rows, columns = tuple(map(int, console_size.strip().split()))
    except ValueError:
        pass
    return dict(rows=rows, columns=columns)


class PIPE_OUT:
    """Pipe stdout/stderr into a different handlers."""

    def __enter__(self):
        return self

    def __init__(self, *handlers: Union[IO[bytes], TextIO]):

        self.handlers = handlers

    def write(self, msg, *args, **kwargs) -> None:
        for handler in self.handlers:
            handler.write(msg, *args, **kwargs)
            # name = handler.name.lower()
            # if name != "<stdout>" and name != "<stderr>":
            #    handler.flush()

    def flush(self) -> None:
        for handler in self.handlers:
            handler.flush()

    def __exit__(self, *args, **kwargs) -> None:
        return None


def supermakedirs(path, mode):
    """
    This snippet of code was taken from stackoverflow.com
    On some systems the parameter for the access rights are ignored when
    using os.makedirs.
    This routine overcomes this problem.
    """
    # this is a neccessary condition,
    # otherwise the path will be created twice
    try:
        if path[-1] == "/":
            path = path[:-1]
    except IndexError:
        pass

    if not path or os.path.exists(path):
        return []
    (head, tail) = os.path.split(path)
    res = supermakedirs(head, mode)
    # In some rare cases we run into race conditions creating directories.
    # This prevents the system from raising an error
    try:
        os.mkdir(path)
        os.chmod(path, mode)
    except OSError as exc:
        if exc.errno != errno.EEXIST:
            raise
        pass
    res += [path]
    return res


def mp_wrap_fn(args):
    """
    Wrapper for multi-processing functions with more than one argument.
    :type args: tuple
    :param args: the function name and its arguments
    """

    function_to_call = args[0]
    args = args[1:]

    return function_to_call(*args)


class Struct:
    """This class is used for converting dictionaries into classes in order
        to access them
    more comfortably using the dot syntax."""

    def __init__(self, **entries):
        self.__dict__.update(entries)

    def __getattr__(self, name):
        return None

    def validate(self, value):
        """Check the value is one of the possibilities from those stored
                in the object.

        :param str_value: the value to check."""
        return value in self.__dict__.values()

    def toDict(self):
        """Transfrom this struct into a dictionary."""
        result = {}
        for i in self.__dict__:
            if isinstance(self.__dict__[i], Struct):
                result[i] = self.__dict__[i].toDict()
            else:
                result[i] = self.__dict__[i]
        return result

    def __repr__(self):
        def to_str(val):
            if isinstance(val, str):
                return "'" + val + "'"
            return val

        return "<%s>" % ",".join(
            [
                "%s:%s" % (att, to_str(self.__dict__[att]))
                for att in self.__dict__
                if not att.startswith("_")
            ]
        )

    def __eq__(self, other):
        if isinstance(other, Struct):
            return self.toDict().__eq__(other.toDict())

    def __hash__(self):
        return 0

    @staticmethod
    def from_dict(dictionary, recurse=False):
        """Transforms a dict into an object.

        :param dictionary: the dictionary that will be transformed.
        :param recurse: if it would be applied recursive to other internal dictionaries.
        :type recurse: bool"""
        dictionary = copy.deepcopy(dictionary)

        # we don't need to recurse if this is not a non-empty iterable sequence
        if not dictionary:
            return dictionary

        # If a list, apply to elements within
        if type(dictionary) == list:
            return list(map(lambda d: Struct.from_dict(d, recurse), dictionary))
        # not a dictionary, return unchanged
        if type(dictionary) != dict:
            return dictionary
        if recurse:
            for key in dictionary:
                dictionary[key] = Struct.from_dict(dictionary[key], recurse)

        return Struct(**dictionary)


class TemplateDict(object):
    """Help object for resolving dictionaries containing substitutable values.
    This object has a *basis* dictionary and a resolution dictionary. The only
    difference among them is that the *basis* dictionary is defined on creation
    and the resolution dictionary is provided when calling the substitution.
    For everything else they work alike and both may contain parameterless functions
    instead of simple values that will get called when resolving them.

    This object uses the :py:class:`string.Template` object for performing
    substitution so the syntax used for substitution is as defined there
    (e.g. ``${var_to_replace}$another_var_to_replace``)."""

    translation_dict = None
    """Stores the translation dictionary used every time
    :class:`TemplateDict.substitute` is call."""

    def __init__(self, **translation_dict):
        """Create a dictionary from the named arguments passed along::

            TempleDict(a='now: $b', b=lambda: datetime.now(), c=lambda: sys.env['PWD'])

        This dictionary is stored at :class:`TemplateDict.translation_dict`."""
        self.translation_dict = translation_dict

    def __wrapDict(self, var_dict={}):
        """Creates a wrapper that acts like a dictionary to the outside but
                performs some operations
        in the background when retrieving the values."""
        templ_self = self

        def f(self, key):

            if key in var_dict:
                val = var_dict[key]
            else:
                val = templ_self.translation_dict[key]
            if callable(val):
                return val()
            else:
                return val

        def i(self):
            for key in list(var_dict.keys()) + list(templ_self.translation_dict.keys()):
                yield (key, self[key])

        return type("dict_wrapper", (object,), {"__getitem__": f, "items": i})()

    def substitute(
        self, substitute_dict: dict[str, Any], recursive: bool = True
    ) -> dict[str, Any]:
        """Substitute the values from substitute dictionary. Values in
                ``substitute_dict`` take precedence from those in
                :class:`TemplateDict.translation_dict`.

        :type recursive: bool
        :param recursive: if the substitution should be resolved recursive by
        including the given
        :type substitute_dict: dict
        :param substitute_dict: is a dictionary of the form: ``variable -> value``.

        ``value`` in ``substitute_dict`` may be any of:
             * a simple value (int, float, str, object).
             * a string containing ``$`` characters as start marks for variables
             which must exists in either ``substitute_dict``
             or :class:`TemplateDict.translate_dict`.
             * a parameterless function returning any of the previous values.

        For instance, given this setup::

            from evaluation_system.misc.utils import TemplateDict
            from time import time
            my_dict = TemplateDict(A='Something: $B',
                                   B='milliseconds ($C)',
                                   C=lambda: '%.12f' % (time() * 1000))
            to_resolve = dict(c='$A', d='The Time in $e', e=lambda: '$B')

        Example:
        --------

            my_dict.substitute(to_resolve)
            {'c': 'Something: milliseconds (1358779343538.392089843750)',
            'e': <function <lambda> at 0xb6f0d374>,
            'd': 'The Time in milliseconds (1358779343538.410888671875)'}

        As you can see functions remain as functions in the ``substitute_dict``
        but they are used for resolution.
        Though there's no guarantee in which order resolution takes place, so
        there's no guarantee that functions are called only once. For instance
        in the example above you see two different calls to :py:func:`time.time`.
        The complete resolution was:

        * For **c**: c = '$A' -> 'Something: $B' -> 'Something: milliseconds ($C)'
                       -> 'Something: milliseconds (1358779343538.392089843750)'
        * For **d**: d = 'The time in $e' -> 'The time in %s' % e()
                       -> 'The time in $B'
                       -> 'The Time in milliseconds (1358779343538.410888671875)'
        Example:
        --------
            my_dict.substitute(to_resolve, recursive=False)
            {'c': 'Something: milliseconds (1358779553514.358886718750)',
            'e': <function <lambda> at 0xb6f0d374>,
            'd': 'The Time in $e'}

        As you see the recursion on :class:`TemplateDict.translation_dict`
        is not affected, but the variables from ``substitute_dict`` are not
        used for substitution at all."""
        # we need to work in a copy if using recursion. But we do this anyways
        # to keep the code simple.
        result = substitute_dict.copy()

        if recursive:
            final_dict = self.__wrapDict(result)
        else:
            # cannot reference itself
            final_dict = self.__wrapDict()

        # accept a maximal recursion of 15 for resolving all tokens
        # 15 is a definite number larger than any thinkable recursion for this case
        max_iter = 15
        recursion = True  # just a mark to know if it's worth iterating at all
        while recursion and max_iter > 0:
            recursion = False  # assume no recursion until one possible case is found
            for var, value in result.items():

                tmpl = None
                if isinstance(value, str) and "$" in value:
                    # something that might need to get replaced!
                    tmpl = Template(value)
                elif isinstance(value, Template):
                    tmpl = value

                if tmpl:
                    result[var] = tmpl.safe_substitute(final_dict)
                    recursion = True
                else:
                    result[var] = value

            max_iter -= 1
        if recursive and max_iter <= 0:
            raise Exception(
                "maximum recursion depth exceeded."
                + "Check the substitution variables are not referencing in a loop.\n"
                + "last state: %s" % ["%s=%s," % (k, v) for k, v in final_dict.items()]
            )

        return result


def find_similar_words(word: str, list_of_valid_words: Iterable[str]) -> List[str]:
    """This function implements a "Did you mean? xxx" search algorith.

    It is used for helping the user find the right word.

    Parameters:
    -----------
    word:
        the word the user selected.
    list_of_valid_words:
        a list of valid words.
    Returns:
    --------
    list : a list of words that are close to the word given"""
    expand_list = {}
    for w in list_of_valid_words:
        for w_part in split(r"[ _\-:/]", w):
            if w_part not in expand_list:
                expand_list[w_part] = set([w])
            else:
                expand_list[w_part].add(w)
    result = get_close_matches(word, expand_list)
    return [w for parts in result for w in expand_list[parts]]


class metadict(dict):
    """A dictionary extension for storing metadata along with the keys.
    In all other cases, it behaves like a normal dictionary."""

    def __init__(self, *args, **kw):
        """Creates a metadict dictionary.
        If the keyword ``compact_creation`` is used and set to ``True`` the
        entries will be given like this:

            key1=(value1, dict1) or key2=value2

        Where dict1 is the dictionary attached to the key providing its
        meta-data (key2 has no meta-data, by the way)."""
        self.metainfo = {}
        compact_creation = kw.pop("compact_creation", False)
        if compact_creation:
            # separate the special "default" in the first field from the
            # dictionary in the second
            super(metadict, self).__init__()
            for key, values in kw.items():
                if isinstance(values, tuple):
                    if len(values) != 2:
                        raise AttributeError(
                            (
                                "On compact creation a tuple with only 2 values is"
                                " expected: (default, metadata)"
                            )
                        )
                    if not isinstance(values[1], dict):
                        raise AttributeError("metadata entry must be a dictionary")
                    self[key] = values[0]
                    self.metainfo[key] = values[1]
                else:
                    self[key] = values
        else:
            super(metadict, self).__init__(*args, **kw)

    def copy(self):
        """:return: a deep copy of this metadict."""
        return deepcopy(self)

    def getMetadata(self, key):
        """Meta-data value associated with this key."""
        if key in self.metainfo:
            return self.metainfo[key]
        else:
            return None

    def setMetadata(self, key, **meta_dict):
        """Store/replace the meta-data allocated for the given key.

        :raises: KeyError if key is not present."""
        if key not in self:
            raise KeyError(key)
        if key not in self.metainfo:
            self.metainfo[key] = {}
        self.metainfo[key].update(meta_dict)

    def clearMetadata(self, key):
        """Clear all meta-data allocated under the given key.

        :raises: KeyError if key is not present."""
        if key not in self:
            raise KeyError(key)
        if key in self.metainfo:
            del self.metainfo[key]

    def put(self, key, value, **meta_dict):
        """Puts a key,value pair into the dictionary and all other keywords are added
        as meta-data to this key. If key was already present, it will be
        over-written and its meta-data will be removed
        (even if no new meta-data is provided)."""
        self[key] = value
        if meta_dict:
            self.clearMetadata(key)
            self.setMetadata(key, **meta_dict)

    @staticmethod
    def hasMetadata(some_dict, key=None):
        """if the given dictionary has meta-data for the given key or, if no key was given,
         if the dictionary can hold meta-data at all.

        Returns:
        --------
         bool: bool
            if ``some_dict`` has stored meta-data for ``key`` or
            any meta-data at all if ``key==None``."""
        if key is None:
            return hasattr(some_dict, "getMetadata")
        else:
            return hasattr(some_dict, "getMetadata") and bool(
                some_dict.getMetadata(key)
            )

    @staticmethod
    def getMetaValue(some_dict, key, meta_key):
        """This method allows to work both with normal dictionaries and metadict transparently.

        :returns: the meta-data associated with the key if any or None if not found or
                  this is not a metadict at all."""
        if metadict.hasMetadata(some_dict):
            meta = some_dict.getMetadata(key)
            if meta and meta_key in meta:
                return meta[meta_key]


class PrintableList(list):
    """
    Helper class which overwrites the __str__ function of list objects
    """

    def __init__(self, *args, **kwargs):
        try:
            self.seperator = kwargs.pop("seperator")
        except KeyError:
            self.seperator = ","
        super(PrintableList, self).__init__(*args, **kwargs)

    def __str__(self):  # pragma: no cover
        """
        :returns: String with comma separated list entries
        """
        return self.seperator.join(map(str, self))

    def __unicode__(self):  # pragma: no cover
        return self.__str__()


class initOrder(object):
    """
    Objects derived by this class get a natural order in the way they are
    initialized. This works especially for several class variables.
    """

    __counter = 0

    def __init__(self):
        self.__number = initOrder.__counter
        initOrder.__counter += 1

    def initCompare(self, other):
        return self.__number - other.__number



"""
Created on 23.05.2016

@author: Sebastian Illing
"""

from datetime import datetime, timedelta
import os
import tempfile
import socket
import shutil
from getpass import getuser

import pytest


def test_store_history(dummy_history, temp_user, dummy_plugin, config_dict):
    row_id = temp_user.getUserDB().storeHistory(
        dummy_plugin, config_dict, "user", 1, caption="My caption"
    )
    h = dummy_history.objects.get(id=row_id)
    assert h
    assert h.status_name() == "finished_no_output"
    assert h.caption == "My caption"
    assert h.config_dict() == config_dict


def test_schedule_entry(dummy_user, dummy_history):
    dummy_user.user.getUserDB().scheduleEntry(
        dummy_user.row_id, dummy_user.username, "/slurm/output/file.txt"
    )
    h = dummy_history.objects.get(id=dummy_user.row_id)
    assert h.status == dummy_history.processStatus.scheduled
    assert h.slurm_output == "/slurm/output/file.txt"
    assert h.host == socket.gethostbyname(socket.gethostname())


def test_upgrade_status(dummy_user, dummy_history):

    with pytest.raises(dummy_user.user.getUserDB().ExceptionStatusUpgrade):
        dummy_user.user.getUserDB().upgradeStatus(
            dummy_user.row_id, dummy_user.username, 6
        )

    dummy_user.user.getUserDB().upgradeStatus(
        dummy_user.row_id, dummy_user.username, dummy_history.processStatus.finished
    )
    h = dummy_history.objects.get(id=dummy_user.row_id)
    assert h.status == dummy_history.processStatus.finished


def test_change_flag(dummy_user, dummy_history):
    dummy_user.user.getUserDB().changeFlag(
        dummy_user.row_id, dummy_user.username, dummy_history.Flag.deleted
    )
    h = dummy_history.objects.get(id=dummy_user.row_id)
    assert h.flag == dummy_history.Flag.deleted


def test_get_history(dummy_user, dummy_plugin, config_dict):
    # create some values
    users = ["user", "other", "user", "test"]
    for u in users:
        dummy_user.user.getUserDB().storeHistory(
            dummy_plugin, config_dict, u, 1, caption=f"My {u}"
        )

    history = dummy_user.user.getUserDB().getHistory()
    assert history.count() == 5
    history = dummy_user.user.getUserDB().getHistory(uid="user")
    assert history.count() == 2
    history = dummy_user.user.getUserDB().getHistory(
        uid="user", tool_name="dummyplugin", limit=2
    )
    assert history.count() == 2
    history = dummy_user.user.getUserDB().getHistory(
        uid=dummy_user.username, entry_ids=dummy_user.row_id
    )
    assert history.count() == 1


def test_add_history_tag(dummy_user, dummy_history):
    from evaluation_system.model.history.models import HistoryTag

    dummy_user.user.getUserDB().addHistoryTag(
        dummy_user.row_id, HistoryTag.tagType.note_public, "Some note"
    )

    h = dummy_history.objects.get(id=dummy_user.row_id)
    tags = h.historytag_set.all()
    assert len(tags) == 1
    assert tags[0].type == HistoryTag.tagType.note_public


def test_update_history_tag(dummy_user):
    from evaluation_system.model.history.models import History, HistoryTag

    dummy_user.user.getUserDB().addHistoryTag(
        dummy_user.row_id, HistoryTag.tagType.note_public, "Some note", uid="user"
    )
    h_tag = History.objects.get(id=dummy_user.row_id).historytag_set.last()
    dummy_user.user.getUserDB().updateHistoryTag(
        h_tag.id, HistoryTag.tagType.note_deleted, "New text", uid="user"
    )
    h_tag = History.objects.get(id=dummy_user.row_id).historytag_set.last()
    assert h_tag.type == HistoryTag.tagType.note_deleted
    assert h_tag.text == "New text"


def test_store_results(dummy_user, dummy_history):

    from evaluation_system.model.history.models import ResultTag

    results = {
        "/some/result.png": {"type": "plot", "caption": "super plot"},
        "/some/other.eps": {"type": "data"},
    }
    dummy_user.user.getUserDB().storeResults(dummy_user.row_id, results)

    h = dummy_history.objects.get(id=dummy_user.row_id)

    assert h.result_set.count() == 2
    for key, val in results.items():
        assert h.result_set.filter(
            history_id_id=dummy_user.row_id, output_file=key
        ).exists()
        if val.get("caption", None):
            res_tag = h.result_set.get(output_file=key).resulttag_set.first()
            assert res_tag.type == ResultTag.flagType.caption
            assert res_tag.text == val["caption"]


def test_version(dummy_user):
    from evaluation_system.model.plugins.models import Version

    Version.objects.all().delete()
    # create version entry
    version_id = dummy_user.user.getUserDB().newVersion(
        "dummyplugin", "1.0", "git", "git_number", "tool_git", "tool_git_number"
    )
    assert Version.objects.filter(id=version_id).exists()

    # get version entry
    get_version_id = dummy_user.user.getUserDB().getVersionId(
        "dummyplugin", "1.0", "git", "git_number", "tool_git", "tool_git_number"
    )
    assert version_id == get_version_id
    Version.objects.all().delete()


def test_create_user(dummy_user):
    from django.contrib.auth.models import User

    User.objects.filter(username="new_user").delete()
    dummy_user.user.getUserDB().createUser("new_user", "test@test.de", "Test", "User")
    assert User.objects.filter(username="new_user").exists()
    User.objects.filter(username="new_user").delete()


def test_create_user_crawl(dummy_user):
    from evaluation_system.model.solr_models.models import UserCrawl
    from django.contrib.auth.models import User

    dummy_user.user.getUserDB().createUser("new_user", "test@test.de", "t", "u")
    dummy_user.user.getUserDB().create_user_crawl("/some/test/folder", "new_user")
    assert UserCrawl.objects.filter(
        status="waiting", path_to_crawl="/some/test/folder"
    ).exists()
    UserCrawl.objects.all().delete()
    User.objects.filter(username="new_user").delete()


def test_timestamp_to_string():
    from evaluation_system.model.db import timestamp_to_string, timestamp_from_string

    time = datetime.now()
    assert timestamp_to_string(time) == time.strftime("%Y-%m-%d %H:%M:%S.%f")


def test_timestamp_from_string():
    from evaluation_system.model.db import timestamp_to_string, timestamp_from_string

    time = datetime.now()
    time_str = timestamp_to_string(time)
    assert time == timestamp_from_string(time_str)



import os
import errno
import threading
import logging
import grp
import socket

from stat import S_IFDIR, S_IFREG
from time import time, sleep, mktime
from datetime import datetime
from subprocess import Popen, PIPE, STDOUT
from errno import *

from fuse import FUSE, FuseOSError, Operations, LoggingMixIn

from evaluation_system.model.esgf import P2P
from evaluation_system.misc import config
from esgf_crawl_config import Solr2EsgfConfig

from cdo import Cdo

logging.getLogger().setLevel(logging.INFO)


class EsgfFuse(Operations):
    def __init__(self):

        self.logcache = config.get("esgf_logcache")
        # self.logcache= '/pf/b/b324029/cache'
        self.esgftmp = "%s/%s/" % (self.logcache, "ESGF_CACHE")
        self.logpath = "%s/%s/" % (self.logcache, "ESGF_LOG")
        self.esgf_server = config.get("esgf_server").split(",")
        self.hostname = socket.gethostname()

        self.certs = config.get("private_key")
        self.wget = config.get("wget_path")

        self.threadLimiter = threading.BoundedSemaphore(
            int(config.get("parallel_downloads"))
        )
        self.rwlock = threading.Lock()

        self.gid = grp.getgrnam("bmx828").gr_gid
        try:
            os.makedirs(self.logpath, 0e0775)
        except OSError as exception:
            if exception.errno != errno.EEXIST:
                raise

    #         try:
    #             file(self.logpath+self.wgetlog,'a+').close()
    #             os.chown(self.logpath+self.wgetlog, -1,self.gid)
    #             os.chmod(self.logpath+self.wgetlog,S_IREAD|S_IWRITE|S_IRGRP|S_IWGRP)
    #
    #             file(self.logpath+self.downlog,'a+').close()
    #             os.chown(self.logpath+self.downlog, -1,self.gid)
    #             os.chmod(self.logpath+self.downlog,S_IREAD|S_IWRITE|S_IRGRP|S_IWGRP)
    #         except IOError as e:
    #             pass

    def get_url(self, path, size=False):

        esgfpath, filename = os.path.split(path)

        if size:
            try:
                size = os.path.getsize(self.esgftmp + path)
                return None, size
            except OSError:
                pass
        try:
            (
                project,
                product,
                institute,
                model,
                experiment,
                time_frequency,
                realm,
                variable,
                ensemble,
            ) = esgfpath[1:].split("/")
        except ValueError:
            raise FuseOSError(ENOENT)

        cmor_path = Solr2EsgfConfig().project_select(esgfpath, filename)
        fields = ["url", "size", "timestamp"]
        print(path)
        facets = {
            "project": cmor_path["project"],
            "type": "File",
            "product": cmor_path["product"],
            "institute": cmor_path["institute"],
            "experiment": cmor_path["experiment"],
            "time_frequency": cmor_path["time_frequency"],
            "realm": cmor_path["realm"],
            "variable": cmor_path["variable"],
            "ensemble": cmor_path["ensemble"],
            "title": cmor_path["filename"],
        }
        timestamp = 0
        url = None

        for server in self.esgf_server:
            p2p = P2P(node=server)
            for ncfile in p2p.get_datasets(fields=",".join(fields), **facets):
                url = [url for url in ncfile["url"] if "application/netcdf" in url][0]
                if url is None:
                    continue
                if (
                    mktime(
                        (
                            datetime.strptime(ncfile["timestamp"], "%Y-%m-%dT%H:%M:%SZ")
                        ).timetuple()
                    )
                    >= timestamp
                ):
                    timestamp = mktime(
                        (
                            datetime.strptime(ncfile["timestamp"], "%Y-%m-%dT%H:%M:%SZ")
                        ).timetuple()
                    )
                    size = int(ncfile["size"])
            if url is not None:
                break
        if url is None:
            raise FuseOSError(ENETUNREACH)
        # raise FuseOSError(ENETUNREACH)
        url = url.split("|")[0]
        try:
            return url, size
        except UnboundLocalError:
            print("PATH: " + path)
            print("No url for this PATH")

    def download(self, esgfpath, path, filename, cdo=Cdo()):
        httppath, _ = self.get_url(path)

        with open(self.esgftmp + path + ".lock", "w"):
            pass
        cmor_path = Solr2EsgfConfig().project_select(esgfpath, filename)
        wgetlog = "%s_wget_%s.log" % (
            socket.gethostname(),
            datetime.now().strftime("%Y%m%d"),
        )
        self.rwlock.release()

        command = (
            self.wget
            + " --no-check-certificate -O "
            + self.esgftmp
            + esgfpath
            + "/"
            + filename
            + ".tmp"
            " --secure-protocol=TLSv1 --certificate "
            + self.certs
            + " --private-key "
            + self.certs
            + " "
            + httppath
        )
        if os.path.isfile(self.esgftmp + esgfpath + "/" + filename + ".tmp"):
            return
        self.threadLimiter.acquire()
        process = Popen(command, shell=True, stdin=PIPE, stdout=PIPE, stderr=STDOUT)
        all_res = process.communicate()[0]
        self.rwlock.acquire()
        with open(self.logpath + wgetlog, "a+") as f:
            f.write(all_res + "\n")
        self.rwlock.release()
        #                         if 'OpenSSL: error:' in line:
        #                             download.write('Certificate Error:\n')
        #                             download.write('URL: '+httppath+'\n')
        #                   print process.wait()

        if cmor_path["project"] == "specs":
            # os.rename(self.esgftmp+esgfpath+'/'+filename+'.tmp',self.esgftmp+esgfpath+'/'+filename)
            cdo.shifttime(
                "-365days",
                input="-selvar,"
                + cmor_path["variable"]
                + " "
                + self.esgftmp
                + esgfpath
                + "/"
                + filename
                + ".tmp",
                output=self.esgftmp + esgfpath + "/" + filename,
                options=" -O ",
            )
            os.remove(self.esgftmp + esgfpath + "/" + filename + ".tmp")
        else:
            os.rename(
                self.esgftmp + esgfpath + "/" + filename + ".tmp",
                self.esgftmp + esgfpath + "/" + filename,
            )
        os.remove(self.esgftmp + path + ".lock")
        self.threadLimiter.release()

    def getattr(self, path, fh=None):
        filepath, extension = os.path.splitext(path)
        st = dict(st_mode=(S_IFDIR | 0e0755), st_nlink=2)
        if extension == ".nc":
            try:
                self.threadLimiter.acquire()
                st = dict(
                    st_mode=(S_IFREG | 0e0444), st_size=self.get_url(path, True)[1]
                )
                self.threadLimiter.release()
            except TypeError:
                raise FuseOSError(ENETUNREACH, "Bla")
        st["st_ctime"] = st["st_mtime"] = st["st_atime"] = time()
        return st

    def open(self, path, fh):

        sleep(0.1)
        esgfpath, filename = os.path.split(path)
        try:
            with open(self.esgftmp + path) as testfile:
                pass
            return fh
        except IOError:
            pass

        self.rwlock.acquire()
        try:
            with open(self.esgftmp + path + ".lock") as testfile:
                pass
            self.rwlock.release()
        except IOError:
            try:
                os.makedirs(self.esgftmp + esgfpath, 0e0775)
            except OSError as exception:
                if exception.errno != errno.EEXIST:
                    raise
            self.download(esgfpath, path, filename)

        while True:
            sleep(5)
            if not os.path.isfile(self.esgftmp + path + ".lock"):
                if os.path.isfile(self.esgftmp + path):
                    break
        return fh

    #         sleep(0.1)
    #         # create directories recursively
    #         try:
    #             os.makedirs(self.esgftmp+esgfpath,0775)
    # #             [os.chown(dir, -1, 1000) for root, dirs, _ in os.walk(self.esgftmp+esgfpath) for dir in dirs]
    # #             [os.chmod(os.path.join(root,dir),S_IREAD|S_IWRITE|S_IRGRP|S_IWGRP) for root, dirs, _ in os.walk(self.esgftmp+esgfpath) for dir in dirs]
    #         except OSError as exception:
    #             if exception.errno != errno.EEXIST: raise
    #         try: # look for lock file - lock file = signal for download
    #                  with open(self.esgftmp+path+'.lock') as testfile:pass
    #                  sleep (.1)
    #                  self.open(path, fh)
    #         except IOError as e:
    #             try:
    #                 with open(self.esgftmp+path) as testfile: pass
    #                 return fh
    #             except IOError as e:
    #                 with self.rwlock:
    #                     self.download(esgfpath,path,filename)
    #                     return fh

    def read(self, path, length, offset, fh):
        with open(self.esgftmp + path) as f:
            f.seek(offset, 0)
            buf = f.read(length)
            f.close()
            return buf

    readdir = None
    access = None
    flush = None
    getxattr = None
    listxattr = None
    opendir = None
    releasedir = None
    statfs = None
    release = None



"""
Created on 30.05.2016

@author: Sebastian Illing
"""
import json
from pathlib import Path
import shlex

from evaluation_system.tests import run_cli


def test_show_facet(capsys, dummy_config):

    run_cli("esgf --show-facet=project,product -d")
    res = capsys.readouterr().out
    assert "[product]" in res
    assert "[project]" in res


def test_query(capsys, search_dict, dummy_config):
    from freva.cli.esgf import main as run

    run_cli("esgf --debug project=TEST --query=project,product --opendap")
    res = capsys.readouterr().out
    res = json.loads(res)[0]
    assert "project" in res.keys()
    run_cli("esgf -d project=TEST limit=1 --query=project --gridftp")
    res = capsys.readouterr().out
    assert "['TEST']" in res
    run(shlex.split("project=TEST --query=project --gridftp"))
    res = capsys.readouterr().out
    res = [r for r in res.split("\n") if r.strip()]
    num_res = len(res)
    run_cli("esgf project=TEST offset=1 --query=project --gridftp")
    res = capsys.readouterr().out
    res = [r for r in res.split("\n") if r.strip()]
    assert num_res == len(res) + 1
    run(["project=TESTs"])
    assert not capsys.readouterr().out
    run(shlex.split("project=TEST --datasets"))
    assert "- version:" in capsys.readouterr().out
    run(["--show-facet=blabla"])
    assert not capsys.readouterr().out


def test_freva_esgf_method(dummy_config):
    from freva import esgf

    result_to_be = [
        "output1/MPI-M/MPI-ESM-LR/historical/day/atmos/day/r1i1p1/v20111006/"
        "tas/tas_day_MPI-ESM-LR_historical_r1i1p1_18500101-18591231.nc",
        "output1/MPI-M/MPI-ESM-LR/historical/day/atmos/day/r1i1p1/v20111006/"
        "tas/tas_day_MPI-ESM-LR_historical_r1i1p1_18600101-18691231.nc",
    ]
    res = " ".join(
        esgf(
            project="CMIP5",
            experiment="historical",
            variable="tas",
            institute="MPI-M",
            time_frequency="day",
        )
    )
    for f in result_to_be:
        assert f in res
    res = esgf(show_facet="product")
    res = res["product"]["MRE2reanalysis"]
    assert res == 6
    fn = Path("/tmp/file_script.sh")
    res = esgf(
        project="CMIP5",
        model="MPI-ESM-LR",
        experiment="decadal2001",
        variable="tas",
        distrib=False,
        download_script=fn,
    )
    assert fn.is_file()
    assert res == f"Download script successfully saved to {fn}"
    assert oct(fn.stat().st_mode)[-3:] == "755"


def test_find_files(capsys, search_dict, dummy_config):

    result_to_be = [
        "output1/MPI-M/MPI-ESM-LR/decadal2000/mon/atmos/Amon/r1i1p1/tas/1/"
        "tas_Amon_MPI-ESM-LR_decadal2000_r1i1p1_200101-201012.nc",
        "output1/MPI-M/MPI-ESM-LR/decadal2000/mon/atmos/Amon/r1i1p1/v20120529/"
        "tas/tas_Amon_MPI-ESM-LR_decadal2000_r1i1p1_200101-201012.nc",
        "output1/MPI-M/MPI-ESM-LR/decadal2000/mon/atmos/Amon/r1i1p1/v20120529/"
        "tas/tas_Amon_MPI-ESM-LR_decadal2000_r1i1p1_200101-201012.nc",
        "output1/MPI-M/MPI-ESM-LR/decadal2000/mon/atmos/Amon/r1i1p1/v20120529/"
        "tas/tas_Amon_MPI-ESM-LR_decadal2000_r1i1p1_200101-201012.nc",
    ]

    run_cli("esgf --show-facet=blabla")
    assert not capsys.readouterr().out
    run_cli(["esgf"] + [f"{key}={val}" for key, val in search_dict.items()])
    res = capsys.readouterr().out
    for f in result_to_be:
        assert f in res
    run_cli(
        ["esgf"] + ["--datasets"] + [f"{key}={val}" for key, val in search_dict.items()]
    )
    res = capsys.readouterr().out
    assert (
        "cmip5.output1.MPI-M.MPI-ESM-LR.decadal2000.mon.atmos.Amon.r1i1p1 - version: 20120529"
        in res
    )


def test_download_script(capsys, search_dict, tmp_dir, dummy_config):

    fn = tmp_dir / "download_test_script.sh"
    run_cli(
        ["esgf"]
        + ["%s=%s" % (key, val) for key, val in search_dict.items()]
        + [f"--download-script={fn}"]
    )
    res = capsys.readouterr().out
    assert fn.is_file()
    assert res == f"Download script successfully saved to {fn}\n"
    assert oct(fn.stat().st_mode)[-3:] == "755"



import shlex
import pytest

from evaluation_system.tests import run_cli


def test_index_len(dummy_solr):
    assert dummy_solr.all_files.status()["index"]["numDocs"] == 5
    assert dummy_solr.latest.status()["index"]["numDocs"] == 3


def test_freva_databrowser_method(dummy_solr):
    from freva import databrowser

    all_files_output = sorted(
        [
            f"{dummy_solr.tmpdir}/cmip5/output1/MOHC/HadCM3/historical/mon/aerosol/aero/r2i1p1/v20110728/wetso2/wetso2_aero_HadCM3_historical_r2i1p1_190912-193411.nc",
            f"{dummy_solr.tmpdir}/cmip5/output1/MOHC/HadCM3/decadal2009/mon/atmos/Amon/r7i2p1/v20110819/ua/ua_Amon_HadCM3_decadal2009_r7i2p1_200911-201912.nc",
            f"{dummy_solr.tmpdir}/cmip5/output1/MOHC/HadCM3/decadal2008/mon/atmos/Amon/r9i3p1/v20120523/tauu/tauu_Amon_HadCM3_decadal2008_r9i3p1_200811-201812.nc",
        ]
    )
    res = list(databrowser(variable="ua"))
    assert res == [
        f"{dummy_solr.tmpdir}/cmip5/output1/MOHC/HadCM3/decadal2009/mon/atmos/Amon/r7i2p1/v20110819/ua/ua_Amon_HadCM3_decadal2009_r7i2p1_200911-201912.nc"
    ]
    target = sorted(
        [
            f"{dummy_solr.tmpdir}/cmip5/output1/MOHC/HadCM3/decadal2009/mon/atmos/Amon/r7i2p1/v20110819/ua/ua_Amon_HadCM3_decadal2009_r7i2p1_200911-201912.nc",
            f"{dummy_solr.tmpdir}/cmip5/output1/MOHC/HadCM3/decadal2008/mon/atmos/Amon/r9i3p1/v20120523/tauu/tauu_Amon_HadCM3_decadal2008_r9i3p1_200811-201812.nc",
        ]
    )
    res = sorted(databrowser(variable=["ua", "tauu"]))
    assert res == target
    res = sorted(databrowser(variable=["ua", "tauu", "wetso2"]))
    assert res == all_files_output
    v = "v20110419"
    res = sorted(databrowser(variable="ua", version=v))
    assert v in res[0]
    with pytest.raises(TypeError):
        databrowser("badoption")
    res = databrowser(all_facets=True, relevant_only=True)
    assert isinstance(res, dict)
    target = sorted(
        [
            "cmor_table",
            "product",
            "realm",
            "dataset",
            "institute",
            "project",
            "time_frequency",
            "experiment",
            "variable",
            "model",
            "ensemble",
        ]
    )
    relevant = ["cmor_table", "ensemble", "experiment", "realm", "variable"]
    res = sorted(databrowser(attributes=True))
    assert sorted(target) == res
    res = sorted(databrowser(attributes=True, relevant_only=True))
    assert relevant == res


def test_search_files_cmd(dummy_solr, capsys):
    from evaluation_system.misc.exceptions import CommandError
    from freva.cli.databrowser import main as run

    cmd = shlex.split("databrowser")
    all_files_output = sorted(
        [
            f"{dummy_solr.tmpdir}/cmip5/output1/MOHC/HadCM3/historical/mon/aerosol/aero/r2i1p1/v20110728/wetso2/wetso2_aero_HadCM3_historical_r2i1p1_190912-193411.nc",
            f"{dummy_solr.tmpdir}/cmip5/output1/MOHC/HadCM3/decadal2009/mon/atmos/Amon/r7i2p1/v20110819/ua/ua_Amon_HadCM3_decadal2009_r7i2p1_200911-201912.nc",
            f"{dummy_solr.tmpdir}/cmip5/output1/MOHC/HadCM3/decadal2008/mon/atmos/Amon/r9i3p1/v20120523/tauu/tauu_Amon_HadCM3_decadal2008_r9i3p1_200811-201812.nc",
        ]
    )
    run_cli(cmd)
    res = sorted([f for f in capsys.readouterr().out.split("\n") if f])
    assert len(res) == len(all_files_output)
    assert res == all_files_output
    run(["variable=ua"])
    res = capsys.readouterr().out
    assert (
        res
        == f"{dummy_solr.tmpdir}/cmip5/output1/MOHC/HadCM3/decadal2009/mon/atmos/Amon/r7i2p1/v20110819/ua/ua_Amon_HadCM3_decadal2009_r7i2p1_200911-201912.nc\n"
    )
    res = run_cli(cmd + ["variable=ua", "variable=tauu"])
    res = capsys.readouterr().out
    target = sorted(
        [
            f"{dummy_solr.tmpdir}/cmip5/output1/MOHC/HadCM3/decadal2009/mon/atmos/Amon/r7i2p1/v20110819/ua/ua_Amon_HadCM3_decadal2009_r7i2p1_200911-201912.nc",
            f"{dummy_solr.tmpdir}/cmip5/output1/MOHC/HadCM3/decadal2008/mon/atmos/Amon/r9i3p1/v20120523/tauu/tauu_Amon_HadCM3_decadal2008_r9i3p1_200811-201812.nc",
        ]
    )
    res = sorted([f for f in sorted(res.split("\n")) if f])
    assert res == target
    run_cli(cmd + ["variable=ua", "variable=tauu", "variable=wetso2"])
    res = capsys.readouterr().out
    res = [f for f in sorted(res.split("\n")) if f]
    assert res == all_files_output
    # search specific version
    v = "v20110419"
    run_cli(cmd + ["variable=ua", f"version={v}"])
    res = capsys.readouterr().out
    assert v in res
    # test bad input
    with pytest.raises(CommandError):
        run_cli(cmd + ["badoption"])


def test_search_facets(dummy_solr, capsys):
    cmd = shlex.split("databrowser")
    all_facets = [
        map(str.strip, f.split(":"))
        for f in """cmor_table: aero,amon
product: output1
realm: aerosol,atmos
dataset: cmip5
institute: mohc
project: cmip5
time_frequency: mon
experiment: decadal2008,decadal2009,historical
variable: tauu,ua,wetso2
model: hadcm3
ensemble: r2i1p1,r7i2p1,r9i3p1
""".split(
            "\n"
        )
        if f
    ]
    run_cli(cmd + ["--all-facets"])
    res = capsys.readouterr().out
    res = [map(str.strip, f.split(":")) for f in res.split("\n") if f]
    assert dict(res) == dict(all_facets)
    all_facets = {
        "product": ["output1"],
        "realm": ["aerosol", "atmos"],
        "dataset": ["cmip5"],
        "institute": ["mohc"],
        "project": ["cmip5"],
        "time_frequency": ["mon"],
        "experiment": ["decadal2008", "decadal2009", "historical"],
        "variable": ["tauu", "ua", "wetso2"],
        "model": ["hadcm3"],
        "ensemble": ["r2i1p1", "r7i2p1", "r9i3p1"],
    }
    run_cli(cmd + ["--facet=variable"])
    res = capsys.readouterr().out
    assert res == "variable: tauu,ua,wetso2\n"
    run_cli(cmd + ["--facet=variable", "experiment=historical"])
    res = capsys.readouterr().out
    assert res == "variable: wetso2\n"
    run_cli(cmd + ["--facet=variable", "--facet-limit=2"])
    res = capsys.readouterr().out
    assert res == "variable: tauu,ua,...\n"
    run_cli(cmd + ["--facet=variable", "--count-facet-values"])
    res = capsys.readouterr().out
    assert res == "variable: tauu (1),ua (1),wetso2 (1)\n"


def test_show_attributes(dummy_solr, capsys):

    cmd = shlex.split("databrowser --attributes")
    run_cli(cmd)
    res = capsys.readouterr().out
    res = sorted([f.strip() for f in res.split(",") if f])
    target = sorted(
        [
            "cmor_table",
            "product",
            "realm",
            "dataset",
            "institute",
            "project",
            "time_frequency",
            "experiment",
            "variable",
            "model",
            "ensemble",
        ]
    )
    assert target == res


def test_solr_backwards(dummy_solr, capsys):

    cmd = shlex.split("databrowser --all-facets")
    cmd += [
        f"file={dummy_solr.tmpdir}/cmip5/output1/MOHC/HadCM3/decadal2008/mon/atmos/Amon/r9i3p1/v20120523/tauu/tauu_Amon_HadCM3_decadal2008_r9i3p1_200811-201812.nc"
    ]
    run_cli(cmd)
    res = capsys.readouterr().out
    res = [map(str.strip, f.split(":")) for f in res.split("\n") if f]
    target = [
        map(str.strip, f.split(":"))
        for f in """cmor_table: amon
product: output1
realm: atmos
dataset: cmip5
institute: mohc
project: cmip5
time_frequency: mon
experiment: decadal2008
variable: tauu
model: hadcm3
ensemble: r9i3p1
""".split(
            "\n"
        )
        if f
    ]
    print(list(res))
    assert dict(res) == dict(target)



"""Collection of admin commands to create flat pages."""

from __future__ import annotations
import argparse
import os
from pathlib import Path
import shutil
import shlex
from subprocess import run, PIPE
import sys
from tempfile import TemporaryDirectory
from typing import Any, Optional

from django.contrib.flatpages.models import FlatPage

from ..utils import BaseParser, is_admin

import evaluation_system.api.plugin_manager as pm
from evaluation_system.misc import config, logger

__all__ = ["update_tool_doc"]
CLI = "DocCli"


class Convert2Html:
    """Converter class that converts different file formats to html."""

    def __init__(self, input_file: Path, tmpdir: Path):

        self.tmpdir = tmpdir
        shutil.copytree(input_file.parent, self.tmpdir)
        self.input_file = tmpdir / input_file.name
        self.input_dir = tmpdir
        self.html_file = self.input_file.with_suffix(".html")
        suffix = Path(input_file).suffix.strip(".")
        self.conv_func = getattr(self, f"convert_{suffix}")

    def convert_tex(self) -> str:
        """Convert latex to html files."""
        bibfiles = [str(f) for f in self.input_dir.rglob("*.bib")]
        cmd = f"pandoc {self.input_file} -f latex -t html5"
        if bibfiles:
            cmd += f" --bibliography {bibfiles[0]}"
        cmd += f" -o {self.html_file}"
        return cmd

    def convert(self, tool: str) -> str:
        """Convert the input to html file."""
        cmd = self.conv_func()
        env = os.environ.copy()
        env["PATH"] = f"{Path(sys.exec_prefix) / 'bin'}:{env['PATH']}"
        _ = run(
            shlex.split(cmd),
            stdout=PIPE,
            stderr=PIPE,
            env=env,
            check=True,
            cwd=self.input_file.parent,
        )
        with self.html_file.open() as fi:
            text = fi.read()
        # replace img src
        text = text.replace(
            'src="figures/', 'style="width:80%;" src="/static/preview/doc/' + tool + "/"
        )
        # remove too big sigma symbols
        return text.replace('mathsize="big"', "")


def update_tool_doc(tool_name: str, master_doc: Optional[Path] = None) -> None:
    """Update the html files of tool documentation"""
    is_admin(raise_error=True)
    plugin_path = Path(pm.get_plugin_instance(tool_name).class_basedir or ".")
    doc_file = Path(master_doc or plugin_path / "doc" / "{tool_name}.tex")
    if not doc_file.is_file():
        raise FileNotFoundError(f"No documentation found in {doc_file.parent}")
    tool = tool_name.lower()
    # copy folder to /tmp for processing
    config.reloadConfiguration()
    with TemporaryDirectory(prefix=tool, suffix="_doc") as td:
        conv = Convert2Html(doc_file, Path(td) / "doc")
        html_text = conv.convert(tool)
        flat_page, created = FlatPage.objects.get_or_create(
            title=tool_name, url=f"/about/{tool}/"
        )
        flat_page.content = html_text
        flat_page.save()
        # Copy images to website preview path
        preview_path = Path(config.get("preview_path"))
        dest_dir = preview_path / f"doc/{tool}"
        dest_dir.mkdir(parents=True, exist_ok=True)
        logger.info(f"Flat pages for {tool_name} has been created")


class DocCli(BaseParser):
    """Interface defining parsers for documetation update."""

    desc = "Update the plugin documentation."

    def __init__(self, parser: argparse.ArgumentParser) -> None:
        """Construct the sub arg. parser."""
        self.sub_commands = {}
        parser.add_argument("tool", help="Plugin name", type=str)
        parser.add_argument(
            "--file-name",
            help=(
                "Filename of the main docu file. Standard doc location "
                " is taken if None given (default)."
            ),
            default=None,
            type=Path,
        )
        self.parser = parser
        parser.add_argument(
            "--debug",
            "--verbose",
            help="Use verbose output.",
            action="store_true",
            default=False,
        )
        self.logger.setLevel(20)  # Set log level to info
        self.parser.set_defaults(apply_func=self.run_cmd)

    @staticmethod
    def run_cmd(args: argparse.Namespace, **kwargs: Any) -> None:
        """Apply the check4broken_runs method"""

        update_tool_doc(args.tool, args.file_name)



from evaluation_system.api.workload_manager.slurm import SLURMJob


def test_header():
    with SLURMJob(walltime="00:02:00", processes=4, cores=8, memory="28GB") as cluster:

        assert "#SBATCH" in cluster.job_header
        assert "#SBATCH -J worker" in cluster.job_header
        assert "#SBATCH -n 1" in cluster.job_header
        assert "#SBATCH --cpus-per-task=8" in cluster.job_header
        assert "#SBATCH --mem=27G" in cluster.job_header
        assert "#SBATCH -t 00:02:00" in cluster.job_header
        assert "#SBATCH -p" not in cluster.job_header
        # assert "#SBATCH -A" not in cluster.job_header

    with SLURMJob(
        queue="regular",
        project="DaskOnSlurm",
        processes=4,
        cores=8,
        memory="28GB",
        job_cpu=16,
        job_mem="100G",
    ) as cluster:

        assert "#SBATCH --cpus-per-task=16" in cluster.job_header
        assert "#SBATCH --cpus-per-task=8" not in cluster.job_header
        assert "#SBATCH --mem=100G" in cluster.job_header
        assert "#SBATCH -t " in cluster.job_header
        assert "#SBATCH -A DaskOnSlurm" in cluster.job_header
        assert "#SBATCH -p regular" in cluster.job_header

    with SLURMJob(cores=4, memory="8GB") as cluster:

        assert "#SBATCH" in cluster.job_header
        assert "#SBATCH -J " in cluster.job_header
        assert "#SBATCH -n 1" in cluster.job_header
        assert "#SBATCH -t " in cluster.job_header
        assert "#SBATCH -p" not in cluster.job_header
        # assert "#SBATCH -A" not in cluster.job_header


def test_job_script():
    with SLURMJob(walltime="00:02:00", processes=4, cores=8, memory="28GB") as cluster:

        job_script = cluster.job_script()
        assert "#SBATCH" in job_script
        assert "#SBATCH -J worker" in job_script
        assert "#SBATCH -n 1" in job_script
        assert "#SBATCH --cpus-per-task=8" in job_script
        assert "#SBATCH --mem=27G" in job_script
        assert "#SBATCH -t 00:02:00" in job_script
        assert "#SBATCH -p" not in job_script
        # assert "#SBATCH -A" not in job_script

        assert "export " not in job_script

    with SLURMJob(
        walltime="00:02:00",
        processes=4,
        cores=8,
        memory="28GB",
        env_extra=[
            'export LANG="en_US.utf8"',
            'export LANGUAGE="en_US.utf8"',
            'export LC_ALL="en_US.utf8"',
        ],
    ) as cluster:
        job_script = cluster.job_script()
        assert "#SBATCH" in job_script
        assert "#SBATCH -J worker" in job_script
        assert "#SBATCH -n 1" in job_script
        assert "#SBATCH --cpus-per-task=8" in job_script
        assert "#SBATCH --mem=27G" in job_script
        assert "#SBATCH -t 00:02:00" in job_script
        assert "#SBATCH -p" not in job_script
        # assert "#SBATCH -A" not in job_script

        assert 'export LANG="en_US.utf8"' in job_script
        assert 'export LANGUAGE="en_US.utf8"' in job_script
        assert 'export LC_ALL="en_US.utf8"' in job_script



from __future__ import annotations
import argparse
import sys
from typing import Any, Optional

from evaluation_system import __version__
import freva
from .utils import BaseCompleter, BaseParser

CLI = "DataBrowserCli"


class DataBrowserCli(BaseParser):
    """Class that constructs the Databrowser Argument Parser."""

    desc = "Find data in the system."

    def __init__(
        self,
        command: str = "freva",
        parser: Optional[argparse.ArgumentParser] = None,
    ):
        """Construct the databrwoser sub arg. parser."""
        subparser = parser or argparse.ArgumentParser(
            prog=f"{command}-databrowser",
            description=self.desc,
            formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        )
        subparser.add_argument(
            "--multiversion",
            default=False,
            action="store_true",
            help="Select not only the latest version.",
        )
        subparser.add_argument(
            "--relevant-only",
            default=False,
            action="store_true",
            help="Show only search with results >1 possible values",
        )
        subparser.add_argument(
            "--batch-size",
            default=10,
            type=int,
            help="Number of files to retrieve",
        )
        subparser.add_argument(
            "--count-facet-values",
            default=False,
            action="store_true",
            help="Show the number of files for each values in each facet.",
        )
        subparser.add_argument(
            "--attributes",
            default=False,
            action="store_true",
            help=(
                "Retrieve all possible attributes for the current "
                "search instead of the files."
            ),
        )
        subparser.add_argument(
            "--all-facets",
            default=False,
            action="store_true",
            help=("retrieve all facets (attributes & values) instead of " "the files"),
        )
        subparser.add_argument(
            "--facet",
            default=None,
            type=str,
            action="append",
            help=("Retrieve values of given facet instead of files"),
        )
        subparser.add_argument(
            "--facet-limit",
            type=int,
            help="Limit the number of output facets.",
            default=sys.maxsize,
        )
        subparser.add_argument(
            "--debug",
            "-v",
            "-d",
            "--verbose",
            help="Use verbose output.",
            action="store_true",
            default=False,
        )
        subparser.add_argument(
            "facets", nargs="*", help="Search facet(s)", type=str, metavar="facets"
        )
        self.parser = subparser
        self.parser.set_defaults(apply_func=self.run_cmd)

    @staticmethod
    def run_cmd(
        args: argparse.Namespace,
        other_args: Optional[list[str]] = None,
        **kwargs: Optional[Any],
    ) -> None:
        """Call the databrowser command and print the results."""
        facets: dict[str, Any] = BaseCompleter.arg_to_dict(args.facets, append=True)
        facet_limit = kwargs.pop("facet_limit")
        _ = kwargs.pop("facets")
        for key, values in facets.items():
            if len(values) == 1:
                facets[key] = values[0]
        merged_args: dict[str, Any] = {**kwargs, **facets}
        out = freva.databrowser(**merged_args)
        # flush stderr in case we have something pending
        sys.stderr.flush()
        if isinstance(out, dict):
            # We have facet values as return values
            for att, values in out.items():
                facet_limit = facet_limit or len(values) + 1
                try:
                    keys = ",".join(
                        [
                            f"{k} ({c})"
                            for n, (k, c) in enumerate(values.items())
                            if n < facet_limit
                        ]
                    )
                except AttributeError:
                    keys = ",".join(
                        [v for n, v in enumerate(values) if n < facet_limit]
                    )
                if facet_limit < len(values):
                    keys += ",..."
                print(f"{att}: {keys}", flush=True)
            return
        if args.attributes:
            print(", ".join(out), flush=True)
            return
        for key in out:
            print(str(key), flush=True)


def main(argv: Optional[list[str]] = None) -> None:
    """Wrapper for entry point script."""
    cli = DataBrowserCli("freva")
    cli.parser.add_argument(
        "-V",
        "--version",
        action="version",
        version="%(prog)s {version}".format(version=__version__),
    )
    args = cli.parse_args(argv or sys.argv[1:])
    try:
        cli.run_cmd(args, **cli.kwargs)
    except KeyboardInterrupt:  # pragma: no cover
        print("KeyboardInterrupt, exiting", file=sys.stderr, flush=True)
        sys.exit(130)



"""
.. moduleauthor:: Sebastian Illing / estani

This module manages the loading and access to all plug-ins. It is designed
as a central registration through which plug-ins can be accessed.

Plug-ins are automatically registered if found either in a specific
directory ``<evaluation_system_root_dir>/tools`` or by using the
environmental variable ``EVALUATION_SYSTEM_PLUGINS``. This variable must
point to modules containing classes which implement
:class:`evaluation_system.api.plugin.PluginAbstract`.

The value stored ``EVALUATION_SYSTEM_PLUGINS`` is a list of ``path,package``
pairs separated by ``:``. The path denotes where the source is to be
found, and the package is the module name with the class that implements
the PluginAbstract interface. For example:

EVALUATION_SYSTEM_PLUGINS=/path/to/some/dir/,something.else.myplugin:\
/other/different/path,some.plugin:\
/tmp/test,some.other.plugin
"""
from __future__ import annotations
import atexit
from contextlib import contextmanager
import importlib.machinery
import importlib.util
import inspect
import json
import os
import random
import re
import signal
import shutil
import string
import sys
from dataclasses import dataclass
from datetime import datetime
from multiprocessing import Pool
from pathlib import Path
from types import ModuleType
from typing import (
    Any,
    cast,
    Dict,
    Iterator,
    Optional,
    Sequence,
    TypeVar,
    Union,
)
from typing_extensions import TypedDict

from django.db.models.query import QuerySet
from PIL import Image

from evaluation_system import __version__ as version_api
from evaluation_system.misc import config
from evaluation_system.misc import logger as log
from evaluation_system.misc import utils
from evaluation_system.misc.exceptions import (
    ConfigurationException,
    PluginManagerException,
    ParameterNotFoundError,
)
from evaluation_system.model.history.models import Configuration, History, HistoryTag
from evaluation_system.model.plugins.models import Parameter
from evaluation_system.model.user import User
from .plugin import PluginAbstract

PLUGIN_ENV = "EVALUATION_SYSTEM_PLUGINS"
"""Defines the environmental variable name for pointing to the plug-ins"""

IMAGE_RESIZE_EXCEPTIONS = ["PDF", "HDF5", "GRIB"]
"""Plugin output files of these types will not be resized.

    File type is determined by the file extension and the mapping PIL has in
    `Image.registered_extensions()`.
"""

__plugin_modules_user__: dict[str, dict[str, str]] = {}
"""Plugin module info per user. { user => { module name => filepath } }"""
__plugins_meta_user: dict[str, dict[str, PluginMetadata]] = {}
"""Plugin metadata per user. { user => { plugin name => metadata } } """
__version_cache: dict[str, tuple[str, str]] = {}
"""A dictionary which acts as a cache for the git information to
   reduce hard disk access"""


@dataclass
class PluginMetadata:
    name: str
    plugin_class: str
    plugin_module: str
    description: str
    user_exported: bool
    category: str
    tags: list[str]


@dataclass
class _PluginStateHandle:
    """Provide a handle to set/update the plugin states in the db.

    The handle will be instanciated with a state, a db rowid and a user
    object to identify the user running the plugin. Furthermore os
    termination signal will be captured to be able to set the last state
    of the plugin in the db.

    Parameters
    ----------
    status: int
        Integer representation of the plugin state
    rowid: int
        Unique ID of the plugin
    user: User
        user object holding all user information.
    """

    status: int
    rowid: int
    user: User

    def __post_init__(self):
        """Handle various signals and define what to do on exit of this process."""
        # Make sure the last state of this plugin exist of the process is set
        #
        atexit.register(self._update_plugin_state_in_db)
        # Make sure that the last plugin state gets set during when OS
        # termination signals are sent to this process.
        #
        signal.signal(signal.SIGTERM, self._update_plugin_state_in_db_and_quit)
        signal.signal(signal.SIGHUP, self._update_plugin_state_in_db_and_quit)

    def __enter__(self):

        # Initialize the plugin state the database
        self._update_plugin_state_in_db()
        return self

    def __exit__(self, *args):
        self._update_plugin_state_in_db()

    def _update_plugin_state_in_db(self):
        """Update the state of a plugin in the db."""
        self.user.getUserDB().upgradeStatus(
            self.rowid, self.user.getName(), self.status
        )

    def _update_plugin_state_in_db_and_quit(self, *args):
        """Update the plugin state of a plugin and quit."""
        self._update_plugin_state_in_db()
        print("Recieved termination signal: exiting", file=sys.stderr, flush=True)
        sys.exit(1)


T = TypeVar("T")


def munge(seq: Sequence[T]) -> Iterator[T]:
    """Generator to remove duplicates from a list without changing its order.

    Used to keep sys.path tidy.

    Parameters
    ----------
    seq
        any sequence

    Returns
    -------
    iterator
        a generator returning the same objects in the same sequence but
        skipping duplicates.
    """
    seen = set()
    for item in seq:
        if item not in seen:
            seen.add(item)
            yield item


def reload_plugins(user_name: Optional[str] = None) -> None:
    """Reload all plug-ins.

    Plug-ins are then loaded first from the :class:`PLUGIN_ENV`
    environmental variable and then from the configuration file. This means
    that the environmental variable has precedence and can therefore
    overwrite existing plug-ins (useful for debugging and testing).

    Parameters
    ----------
    user_name
        freva user to load the plugins for, if none, it will load the current user
    """

    user_name = user_name or User().getName()

    __plugin_modules__: dict[str, str] = {}
    __plugins__ = {}
    __plugins_meta: dict[str, PluginMetadata] = {}
    __plugin_modules_user__[user_name] = {}
    __plugins_meta_user[user_name] = {}
    extra_plugins = []
    if os.environ.get(PLUGIN_ENV):
        # now get all modules loaded from the environment
        for path, module_name in plugin_env_iter(os.environ[PLUGIN_ENV]):
            # extend path to be exact by resolving all
            # "user shortcuts" (e.g. '~' or '$HOME')
            path = os.path.abspath(os.path.expandvars(os.path.expanduser(path)))
            if os.path.isdir(path):
                # we have a plugin_imp with defined api
                sys.path.append(path)
                # TODO this is not working like in the previous loop. Though we might
                # just want to remove it, as there seem to be no use for this info...
                __plugin_modules__[module_name] = os.path.join(path, module_name)
                extra_plugins.append(module_name)
            else:
                log.warning("Cannot load %s, directory missing: %s", module_name, path)
    # the same for user specific env variable
    if user_name:  # is it possible for User().getName() to be None?
        if PLUGIN_ENV + "_" + user_name in os.environ:
            # now get all modules loaded from the environment
            for path, module_name in plugin_env_iter(
                os.environ[PLUGIN_ENV + "_" + user_name]
            ):
                # extend path to be exact by resolving all
                # "user shortcuts" (e.g. '~' or '$HOME')
                path = os.path.abspath(os.path.expandvars(os.path.expanduser(path)))
                if os.path.isdir(path):
                    # we have a plugin_imp with defined api
                    sys.path.append(path)
                    # TODO this is not working like in the previous loop. Though we
                    # might just want to remove it, as there seem to be no use for
                    # this info...
                    __plugin_modules__[module_name] = os.path.join(path, module_name)
                    extra_plugins.append(module_name)
                else:
                    log.warning(
                        "Cannot load %s, directory missing: %s", module_name, path
                    )
    # get the tools directory from the current one
    # get all modules from the tool directory
    plugins = list(config.get(config.PLUGINS))
    for plugin_name in plugins:
        try:
            py_dir = config.get_plugin(plugin_name, config.PLUGIN_PYTHON_PATH)
            py_mod = config.get_plugin(plugin_name, config.PLUGIN_MODULE)
        except ConfigurationException:
            log.error(
                f"{config.PLUGIN_PYTHON_PATH} and {config.PLUGIN_MODULE}"
                f" need to be configured for plugin {plugin_name}, check config."
            )
            continue
        if os.path.isdir(py_dir):
            if py_mod in __plugin_modules__:
                file_path = __plugin_modules__[py_mod] + ".py"
                log.warning(
                    "Module '%s' is being overwritten by: %s", py_mod, file_path
                )
            else:
                log.debug("Loading '%s'", plugin_name)
                sys.path.append(py_dir)
                __plugin_modules__[plugin_name] = os.path.join(py_dir, py_mod)
        else:
            log.warning("Cannot load '%s' directory missing: %s", plugin_name, py_dir)

    for plugin_name, plugin_mod in __plugin_modules__.items():

        try:
            loader = importlib.machinery.SourceFileLoader(
                plugin_name, plugin_mod + ".py"
            )
            mod = ModuleType(plugin_name)
            loader.exec_module(mod)
            plugin_class = find_plugin_class(mod)

            # in order to get the properties to resolve to their values
            # instead of give function references, we need to instantiate the class
            instance = plugin_class()

            description_str: str = instance.__short_description__
            category: str = instance.__category__
            tags: list[str] = instance.__tags__
            class_name_str = plugin_class.__name__
        except Exception as e:
            # if this isn't a valid plugin, log a warning but continue
            log.warning(f"Error loading plugin {plugin_name}: {e}")
            continue

        if class_name_str != "" and class_name_str.lower() not in __plugins_meta.keys():
            __plugins_meta[class_name_str.lower()] = PluginMetadata(
                name=class_name_str,
                plugin_class=class_name_str,
                plugin_module=plugin_mod,
                description=description_str,
                user_exported=plugin_name in extra_plugins,
                category=category,
                tags=tags,
            )
            __plugins__[class_name_str] = class_name_str
        elif class_name_str != "":
            log.warning(
                "Default plugin %s is being overwritten by: %s",
                class_name_str,
                __plugins_meta[class_name_str.lower()].plugin_module + ".py",
            )
    sys.path = [p for p in munge(sys.path)]
    __plugin_modules_user__[user_name] = __plugin_modules__
    __plugins_meta_user[user_name] = __plugins_meta


def get_plugins_user() -> dict[str, dict[str, PluginMetadata]]:
    """Get plugins per user

    Returns
    -------
    dict[str, dict[str, PluginMetadata]]
        dictionary holding map of user to plugin data
    """
    return __plugins_meta_user


def get_plugins(user_name: Optional[str] = None) -> dict[str, PluginMetadata]:
    """Get plugins and their metadata for the given user

    It's a dictionary with the ``plugin_name`` in lower case of what
    :class:`get_plugin_dict` returns.  This can be used if the plug-in
    name is unknown.

    Parameters
    ----------
    user_name
        If present, gets plugins for given user. Otherwise gets plugins
        for current user

    Returns
    -------
    dict[str, PluginMetadata]
        dict of plugin name to metadata
    """
    user_name = user_name or User().getName()

    return __plugins_meta_user[user_name]


def get_plugin_metadata(
    plugin_name: str, user_name: Optional[str] = None
) -> PluginMetadata:
    """Return the requested plug-in metadata or raise an exception if not found.

    Parameters
    ----------
    plugin_name
        Name of the plug-in to search for.
    user_name
        If present, looks for the plugin for the given user.
        Otherwise uses current user.

    Returns
    -------
    PluginMetadata
        metadata for the plug-in
    """

    user_name = user_name or User().getName()

    plugin_name = plugin_name.lower()
    if plugin_name not in get_plugins(user_name):
        reload_plugins(user_name)
        if plugin_name not in get_plugins(user_name):
            mesg = "No plugin named: %s" % plugin_name
            similar_words = utils.find_similar_words(
                plugin_name, get_plugins(user_name).keys()
            )
            if similar_words:
                mesg = "%s\n Did you mean this?\n\t%s" % (
                    mesg,
                    "\n\t".join(similar_words),
                )
            raise PluginManagerException(mesg + "\n")
    return get_plugins(user_name)[plugin_name]


def get_plugin_instance(
    plugin_name: str, user: Optional[User] = None
) -> PluginAbstract:
    """Return an instance of the requested plug-in.

    At the current time we are just creating new instances, but this might
    change in the future, so it's *not* guaranteed that the instances are
    *unique*, i.e. they might be re-used and/or shared.

    Parameters
    ----------
    plugin_name
        Name of the plugin to search for.
    user
        User for which this plug-in instance is to be acquired. If not
        given the user running this program will be used.

    Returns
    -------
    PluginAbstract
        An instance of the plug-in. Might not be unique.
    """
    user = user or User()

    # in case we want to cache the creation of the plugin classes.
    plugin = get_plugin_metadata(plugin_name, user.getName())
    spec = importlib.util.spec_from_file_location(
        f"{plugin.plugin_class}", f"{plugin.plugin_module}.py"
    )
    if spec is None:
        raise ImportError(
            f"Could not import {plugin.plugin_class} from {plugin.plugin_module}"
        )
    mod = importlib.util.module_from_spec(spec)
    spec_loader = spec.loader
    if spec_loader is None:
        raise ImportError(
            f"Could not import {plugin.plugin_class} from {plugin.plugin_module}"
        )
    spec_loader.exec_module(mod)
    sys.modules[spec.name] = mod
    return getattr(mod, plugin.plugin_class)(user=user)


def parse_arguments(
    plugin_name: str,
    arguments: list[str],
    use_user_defaults: bool = False,
    user: Optional[User] = None,
    config_file: Optional[str] = None,
    check_errors: bool = True,
) -> dict[str, Any]:
    """Parges arguments to send to a plugin.

    Arguments are sent to an instance of the plugin that will handle the
    parsing. This is why the user is required to be known at this stage.

    Parameters
    ----------
    plugin_name
        Name of the plugin to search for.
    arguments
        it will be parsed by the plug-in (see
        :class:`evaluation_system.api.plugin.parse_arguments`)
    use_user_defaults
        If ``True`` and a user configuration is found, this will be used
        as a default for all non set arguments. So the value will be
        determined according to the first found instance of: argument,
        user default, tool default
    user
        The user for whom this arguments are parsed.
    config_file
        Path to a file from where the setup will read a configuration. If
        None, the default user dependent one will be used. This will be
        completely skipped if ``use_user_defaults`` is ``False``.

    Returns
    -------
    dict[str, Any]
        A dictionary with the parsed configuration.
    """
    plugin_name = plugin_name.lower()
    user = user or User()

    p = get_plugin_instance(plugin_name, user)
    complete_conf = p.__parameters__.parse_arguments(
        arguments, use_defaults=True, check_errors=False
    )
    # if we are using user defaults then load them first
    if use_user_defaults:
        user_config_file = user.getUserToolConfig(plugin_name)
        if os.path.isfile(user_config_file):
            with open(user_config_file, "r") as f:
                complete_conf.update(p.read_configuration(f))
    # now if we still have a config file update what the configuration with it
    if isinstance(config_file, str):
        if config_file == "-":
            # reading from stdin
            complete_conf.update(p.read_configuration(sys.stdin))
        elif config_file is not None:
            with open(config_file, "r") as f:
                complete_conf.update(p.read_configuration(f))
    # update with user defaults if desired
    complete_conf.update(
        p.__parameters__.parse_arguments(arguments, check_errors=False)
    )
    # we haven't check for errors because we might have a half implemented
    # configuration some required field might have already been setup
    # (user/system defaults, files, etc)
    # but better if we check them
    if check_errors:
        p.__parameters__.validate_errors(complete_conf, raise_exception=True)
    return complete_conf


def write_setup(
    plugin_name: str,
    config_dict: Optional[Union[PluginMetadata, dict[str, str]]] = None,
    user: Optional[User] = None,
    config_file: Optional[Union[str, Path]] = None,
) -> Union[str, Path]:
    """Writes the plug-in setup to disk.

    This is the configuration for the plug-in itself and not that of the
    tool (which is what normally the plug-in encapsulates). The plug-in
    is not required to write anything to disk when running the tool; it
    might instead configure it from the command line, environmental
    variables or any other method.

    Parameters
    ----------
    plugin_name
        Name of the referred plugin.
    config_dict
        The configuration being stored. If None, the default configuration
        will be stored, this might be incomplete.
    user
        The user for whom this arguments are parsed.
    config_file
        The path to the file where the setup will be stored. If None, the
        default user dependent one will be used. This will be completely
        skipped if ``use_user_defaults`` is ``False``.

    Returns
    -------
    Union[str, Path]
        The path to the configuration file that was written.
    """
    plugin_name = plugin_name.lower()
    user = user or User()
    cfg = cast(Dict[str, Union[str, int, float, bool, None]], config_dict or {})
    p = get_plugin_instance(plugin_name, user)
    complete_conf = p.setup_configuration(
        config_dict=cfg, check_cfg=False, substitute=False
    )

    if config_file is None:
        # make sure the required directory structure and data is in place
        user.prepareDir()

        config_file = user.getUserToolConfig(plugin_name, create=True)

    if config_file == "-":
        p.save_configuration(sys.stdout, config_dict=complete_conf)
    else:
        with open(config_file, "w") as f:
            p.save_configuration(f, config_dict=complete_conf)
    return config_file


def _preview_copy(source_path: str, dest_path: str) -> None:
    """Copy images for preview.

    If the source file is a recognized image file, it will be resized for
    preview, otherwise it will be copied as is with the exception of
    PDFs which are supported by PIL but previously were excluded from
    being resized.

    Parameters
    ----------
    source_path
        the source
    dest_path
        the destination
    """

    extension = Path(source_path).suffix
    supported = Image.registered_extensions()
    if extension in supported and supported[extension] not in IMAGE_RESIZE_EXCEPTIONS:
        _preview_convert(source_path, dest_path)
    else:
        shutil.copyfile(source_path, dest_path)


def _preview_convert(source_path: str, dest_path: str, width: int = 800) -> None:
    """Resizes and converts image

    Resulting image format is based on dest_path file extension

    Parameters
    ----------
    source_path
        the file name of the file to convert
    dest_path
        the file name of the converted file
    width
        width in pixels of resulting image, aspect ratio is maintained
        so height is calculated automatically. Default is 800
    """

    with Image.open(source_path) as img:
        x, y = img.size
        height = 2 * (int(y / (x / width)) // 2)
        im_resize = img.resize((width, height))
        im_resize.save(dest_path)
    os.chmod(dest_path, 509)


def _preview_generate_name(plugin_name: str, metadata: dict[str, Any]) -> str:
    """Creates a filename for plugin output.

    Creates a unique name according to the plugin_name and an eight
    character random string.

    Parameters
    ----------
    plugin_name
        Name of the referred plugin.
    metadata
        The meta-data for the file, to access timestamp

    Returns
    -------
    str
        Generated filename string
    """
    random_suffix = "".join(random.choice(string.ascii_letters) for i in range(8))
    ctime = metadata.get("timestamp", "")
    if ctime:
        time_string = datetime.fromtimestamp(ctime).strftime("%Y%m%d_%H%M%S")
        ctime = "%s_" % time_string
    return plugin_name + "_" + ctime + random_suffix


def _preview_unique_file(plugin_name: str, ext: str, metadata: dict[str, str]) -> str:
    """Creates a unique filename for the preview

    Parameters
    ----------
    plugin_name
        Name of the referred plugin.
    ext
        The extension of the file to be created
    metadata
        the meta-data for the file, to access timestamp
    """
    path = config.get(config.PREVIEW_PATH)
    subdir = datetime.now().strftime("%Y%m%d")
    name = _preview_generate_name(plugin_name, metadata)
    name += ext
    full_path = os.path.join(path, subdir)
    full_name = os.path.join(full_path, name)
    if path.strip() and not os.path.isdir(full_path):
        utils.supermakedirs(full_path, 0o2777)
    return full_name


def _preview_create(plugin_name: str, result: utils.metadict) -> list[str]:
    """Creates the preview.

    Also adds the created files to the result dictionary.

    Parameters
    ----------
    plugin_name
        Name of the referred plugin.
    meta_dict
        A meta dictionary describing the result files

    Returns
    -------
    list[str]
        List of output files
    """
    todo_list = []
    result_list = []
    for file_name in result:
        metadata = result[file_name]
        todo = metadata.get("todo", "")
        if todo == "copy":
            ext = os.path.splitext(file_name)[-1]
            target_name = _preview_unique_file(plugin_name, ext, metadata)
            todo_list.append((_preview_copy, file_name, target_name))
            metadata["preview_path"] = target_name
            result_list.append(target_name)
        elif todo == "convert":
            target_name = _preview_unique_file(plugin_name, ".png", metadata)
            todo_list.append((_preview_convert, file_name, target_name))
            metadata["preview_path"] = target_name
            result_list.append(target_name)
        result[file_name] = metadata
    preview_path = config.get(config.PREVIEW_PATH)
    if preview_path.strip() and todo_list:
        p = Pool(config.NUMBER_OF_PROCESSES)
        p.map(utils.mp_wrap_fn, todo_list)
    return result_list


def run_tool(
    plugin_name: str,
    config_dict: Optional[Union[dict[str, str], utils.metadict]] = None,
    user: Optional[User] = None,
    scheduled_id: Optional[int] = None,
    caption: Optional[str] = None,
    unique_output: bool = True,
) -> Optional[utils.metadict]:
    """Runs a tool and stores the run information.

    Run information is stored in :class:`evaluation_system.model.db.UserDB`.

    Parameters
    ----------
    plugin_name
        Name of the plugin to run.
    config_dict
        The configuration used for running the tool. If is None, the
        default configuration will be stored, this might be incomplete.
    user
        The user starting the tool
    scheduled_id
        If the process is already scheduled then put the row id here
    caption
        The caption to set.

    Returns
    -------
    Optional[utils.metadict]
        Output from the plugin
    """
    plugin_name = plugin_name.lower()
    user = user or User()
    config_dict = config_dict or {}
    p = get_plugin_instance(plugin_name, user)
    complete_conf: dict[str, Union[str, int, float, bool, None]] = {}
    # check whether a scheduled id is given
    if scheduled_id:
        config_dict = cast(
            Dict[str, str],
            load_scheduled_conf(plugin_name, scheduled_id, user),
        )
    if not config_dict:
        conf_file = user.getUserToolConfig(plugin_name)
        if os.path.isfile(conf_file):
            log.debug("Loading config file %s", conf_file)
            with open(conf_file, "r") as f:
                complete_conf = cast(
                    Dict[str, Union[str, int, float, bool, None]],
                    p.read_configuration(f),
                )
        else:
            log.debug("No config file was found in %s", conf_file)
    if not complete_conf:
        # at this stage we want to resolve or tokens and perform some kind of sanity
        # check before going further
        cfg = cast(
            Dict[str, Union[str, int, float, bool, None]],
            {k: v for (k, v) in config_dict.items()},
        )
        complete_conf = p.setup_configuration(cfg, recursion=True)
    log.debug("Running %s with %s", plugin_name, complete_conf)
    out_file: Optional[Path] = None
    rowid = 0
    if scheduled_id:
        rowid = scheduled_id
        out_file = Path(History.objects.get(pk=scheduled_id).slurm_output)
    else:
        version_details = get_version(plugin_name)
        rowid = user.getUserDB().storeHistory(
            p,
            complete_conf,
            user.getName(),
            History.processStatus.running,
            version_details=version_details,
            caption=caption,
        )
    with _PluginStateHandle(
        rowid=rowid, status=History.processStatus.running, user=user
    ) as plugin_state:
        if user:
            # follow the notes
            follow_history_tag(rowid, user.getName(), "Owner")
        p.rowid = rowid
        # Set last state to broken, before we start the plugin
        # in case something goes wrong it will stay in broken because
        # the PluginStateHandle context manager will set it to the last state
        # regardless (which would be broken).
        plugin_state.status = History.processStatus.broken
        # we want that the rowid to be visible to the tool
        # TODO: not sure if this is really optional, the docs don't say much
        result: Optional[utils.metadict] = p._run_tool(
            config_dict=complete_conf,
            unique_output=unique_output,
            out_file=out_file,
            rowid=rowid,
        )
        # save results when existing
        if result is None:
            plugin_state.status = History.processStatus.finished_no_output
        else:
            # create the preview
            preview_path = config.get(config.PREVIEW_PATH, None)
            if preview_path:
                log.debug("Converting....")
                _preview_create(plugin_name, result)
                log.debug("finished")
            # write the created files to the database
            log.debug("Storing results into data base....")
            user.getUserDB().storeResults(rowid, result)
            log.debug("finished")
            # temporary set all processes to finished
            plugin_state.status = History.processStatus.finished
        return result


def schedule_tool(
    plugin_name: str,
    log_directory: Optional[str] = None,
    config_dict: Optional[dict[str, Optional[Union[str, int, bool, float]]]] = None,
    user: Optional[User] = None,
    caption: Optional[str] = None,
    extra_options: list[str] = [],
    unique_output: bool = True,
) -> tuple[int, str]:
    """Schedules a tool and stores the run information.

    Parameters
    ----------
    plugin_name:
        Name of the plugin to run
    log_directory:
        Directory for the output
    config_dict:
        The configuration used for running the tool. If None, the default
        configuration will be stored, this might be incomplete.
    user:
        The user starting the tool
    scheduled_id:
        If the process is already scheduled then put the row id here
    caption:
        The caption to set.
    extra_options:
        Extra options passed to the workload manager, batchmode only

    Returns:
    -------
    job_id, output_file:
        The slurm job id, path to the std out file.
    """

    plugin_name = plugin_name.lower()
    user = user or User()
    config_dict = config_dict or {}
    p = get_plugin_instance(plugin_name, user)
    complete_conf: dict[str, Union[str, int, float, bool, None]] = {}
    # check whether a scheduled id is given
    if not config_dict:
        conf_file = user.getUserToolConfig(plugin_name)
        if os.path.isfile(conf_file):
            log.debug("Loading config file %s", conf_file)
            with open(conf_file, "r") as f:
                complete_conf = cast(
                    Dict[str, Union[str, int, float, bool, None]],
                    p.read_configuration(f),
                )
        else:
            log.debug("No config file was found in %s", conf_file)
    if not complete_conf:
        # at this stage we want to resolve or tokens and perform some kind of sanity
        # check before going further
        conf = cast(Dict[str, Union[str, int, float, bool, None]], config_dict)
        complete_conf = cast(
            Dict[str, Union[str, int, float, bool, None]],
            p.setup_configuration(conf, recursion=True),
        )
    log.debug("Schedule %s with %s", plugin_name, complete_conf)
    version_details = get_version(plugin_name)
    rowid = user.getUserDB().storeHistory(
        p,
        complete_conf,
        user.getName(),
        History.processStatus.not_scheduled,
        version_details=version_details,
        caption=caption,
    )
    # follow the notes
    follow_history_tag(rowid, user.getName(), "Owner")
    # set the output directory
    if log_directory is None:
        log_directory = user.getUserSchedulerOutputDir()
        log_directory = os.path.join(log_directory, plugin_name)

    if not os.path.exists(log_directory):
        utils.supermakedirs(log_directory, 0o2777)
    # write the std out file
    p.rowid = rowid
    job_id, output_file = p.submit_job_script(
        config_dict=config_dict,
        scheduled_id=rowid,
        user=user,
        log_directory=log_directory,
        unique_output=unique_output,
        extra_options=extra_options,
    )
    # create a standard slurm file to view with less
    with open(output_file, "w") as the_file:
        if job_id:
            the_file.write(
                f"Your job is pending with id {job_id}.\n"
                f"\nThis file was automatically "
                "created by the evaluation system.\n"
                f"It will be overwritten by the output of {plugin_name}.\n"
            )
        else:
            the_file.write(
                f"The job id for the submission of {plugin_name} "
                "could not be retreived.\n"
            )
    # set the slurm output file
    user.getUserDB().scheduleEntry(rowid, user.getName(), output_file)
    return rowid, output_file


def get_history(
    plugin_name: Optional[str] = None,
    limit: int = -1,
    since: Optional[str] = None,
    until: Optional[str] = None,
    entry_ids: list[int] = None,
    user: Optional[User] = None,
) -> QuerySet[History]:
    """Returns the history from the given user.

    This is just a wrapper for the defined db interface accessed via the
    user object. See :class:`evaluation_system.model.db.UserDB.getHistory`
    for more information on this interface.

    Parameters
    ----------
    plugin_name
        Name of plugin to get the history for
    limit
        Limits on number of results to get. -1 means no limit
    since
        Only get results after since.
    until
        Only get results earlier than until.
    entry_ids
        Result entry IDs to filter for.
    user
        User to get plugins results for.

    Returns
    -------
    QuerySet[History]
        Results from the database query.
    """
    user = user or User()

    if plugin_name is not None:
        plugin_name = plugin_name.lower()

    return user.getUserDB().getHistory(
        plugin_name,
        limit,
        since=since,
        until=until,
        entry_ids=entry_ids,
        uid=user.getName(),
    )


def get_command_string(
    entry_id: int,
    user: Optional[User] = None,
    command_name: str = "freva-plugin",
    command_options: str = "",
) -> str:
    """Return the parameter string of a history entry.

    Parameters
    ----------
    entry_id
        The history id.
    user
        A user to access the database

    Returns
    -------
    str
        Command string to run the plugin again with the same config.
    """
    user = user or User()

    h: QuerySet[History] = user.getUserDB().getHistory(entry_ids=int(entry_id))
    return get_command_string_from_row(h[0], command_name, command_options)


class CommandConfig(TypedDict):
    name: str
    options: str
    tool: str
    args: dict[str, Any]


def get_command_config_from_row(
    history_row: History, command_name: str, command_options: str
) -> CommandConfig:
    """Get the configuration of a plugin command.

    Parameters
    ----------
    history_row
        History to get command config from.
    command_name
        Name of the CLI command
    command_options
        Options for the given CLI command

    Returns
    -------
    CommandConfig
        Configuration of given command
    """

    configuration = history_row.config_dict()
    result: CommandConfig = {
        "name": command_name,
        "options": command_options,
        "tool": history_row.tool,
        "args": {},
    }
    # find lists
    re_list_pattern = r"^\[.*\]$"
    re_list = re.compile(re_list_pattern)
    for k, value in configuration.items():
        if value is not None:
            # remove brackets from list if value is string
            if re_list.match(str(value)) and isinstance(value, str):
                value = value[1:-1]
            result["args"][k] = value
    return result


def get_command_string_from_config(
    config: CommandConfig,
) -> str:
    """Get the command string for a command

    Parameters
    ----------
    config
        Command config to get the string for.

    Returns
    -------
    str
        CLI command string to run this command.
    """
    result: list[str] = [
        cast(str, config.get(a, ""))
        for a in ("name", "options", "tool")
        if config.get(a)
    ]
    for key, value in config["args"].items():
        if isinstance(value, list):
            result.append(f"{key}={','.join(value)}")
        elif isinstance(value, bool):
            result.append(f"{key}={str(value).lower()}")
        else:
            result.append(f"{key}={value}")
    return " ".join(result)


def get_command_string_from_row(
    history_row: History,
    command_name: str = "freva-plugin",
    command_options: str = "",
) -> str:
    """Get the command string for a command

    Parameters
    ----------
    history_row
        Histroy entry to get string for.
    command_name
        Name of the CLI command
    command_options
        Options for the given CLI command

    Returns
    -------
    str
        CLI command string to run this command.
    """
    config = get_command_config_from_row(history_row, command_name, command_options)
    return get_command_string_from_config(config)


def load_scheduled_conf(plugin_name: str, entry_id: int, user: User) -> dict[str, str]:
    """Loads the configuration from a scheduled plug-in.

    Parameters
    ----------
    plugin_name
        Name of plugin
    entry_id
        Plugin run's History entry ID
    user
        User that ran the plugin

    Returns
    -------
    dict[str, Any]
        History entry's config dict
    """
    h = get_history(plugin_name=plugin_name, entry_ids=[entry_id], user=user)
    # only one row should be selected
    row = h[0]
    # scheduled jobs only
    if row.status != History.processStatus.scheduled:
        raise Exception("This is not a scheduled job (status %i)!" % row.status)
    return row.config_dict()


def get_config_name(pluginname: str) -> Optional[str]:
    """Returns the name of a tool as written in the configuration file.

    This is especially useful when accessing the configuration. This
    will get plugins for the current user only.

    Parameters
    ----------
    pluginname
        Name of the plugin to get
    """
    import inspect

    try:
        plugin = get_plugin_metadata(pluginname.lower())
        modulename = inspect.getmodulename(plugin.plugin_module + ".py")
        for name, module in __plugin_modules_user__[User().getName()].items():
            if modulename == inspect.getmodulename(module + ".py"):
                return pluginname
    except Exception as e:
        log.debug(e.__str__())
    return None


def get_error_warning(tool_name: str) -> tuple[str, str]:
    """Returns error and warning messages from the config file.

    Parameters
    ----------
    tool_name
        Name of plugin

    Returns
    -------
    tuple[str, str]
        Tuple of (errors, warnings)
    """
    plugin_name = get_config_name(tool_name)
    error_file = ""
    error_message = ""
    warning_file = ""
    warning_message = ""
    try:
        error_file = config.get_plugin(plugin_name, "error_file", "")
        error_message = config.get_plugin(plugin_name, "error_message", "")
        warning_file = config.get_plugin(plugin_name, "warning_file", "")
        warning_message = config.get_plugin(plugin_name, "warning_message", "")
    except ConfigurationException:
        pass
    if error_file:
        try:
            f = open(error_file, "r")
            error_message = f.read()
            f.close()
        except Exception as e:
            if not error_message:
                log.warning("Could not read error description\n%s" % str(e))
                error_message = ""
    if warning_file:
        try:
            f = open(warning_file, "r")
            warning_message = f.read()
            f.close()
        except Exception as e:
            if not warning_message:
                log.warning("Could not read warning\n%s" % str(e))
                warning_message = ""
    error_message = error_message.strip()
    warning_message = warning_message.strip()
    return error_message, warning_message


def follow_history_tag(history_id: int, user_name: str, info: str = "") -> None:
    """Add the history tag follow

    Parameters
    ----------
    history_id
        ID of the history object to tag
    user_name
        User that is following the tag
    info
        information to add to the tag
    """
    tagType = HistoryTag.tagType.follow
    rows = HistoryTag.objects.filter(
        history_id_id=history_id, type=tagType, uid_id=user_name
    )
    if len(rows) == 0:
        user = User(user_name)
        user.getUserDB().addHistoryTag(history_id, tagType, info, uid=user_name)


def unfollow_history_tag(history_id: int, user_name: str) -> None:
    """Unfollow a history tag.

    Updates all follow history tags to unfollow for the specified history
    entry and user.

    Parameters
    ----------
    history_id
        ID of the history object to unfollow
    user_name
        User that is unfollowing the tag
    """
    tagType = HistoryTag.tagType.follow
    rows = HistoryTag.objects.filter(
        history_id_id=history_id, type=tagType, uid_id=user_name
    )
    user = User(user_name)
    for row in rows:
        user.getUserDB().updateHistoryTag(
            row.id, HistoryTag.tagType.unfollow, uid=user_name
        )


def get_plugin_version(pluginname: str) -> tuple[str, str]:
    """Extracts plugin version from plugin module.

    Parameters
    ----------
    pluginname
        name of the plugin

    Returns
    -------
    tuple[str, str]
        Tuple of (repository URL, git hash)
    """
    from inspect import currentframe, getfile

    import evaluation_system.model.repository as repository

    version = __version_cache.get(pluginname, None)
    if version is None:
        plugin = get_plugins().get(pluginname, None)
        srcfile = ""
        if plugin is not None:
            srcfile = plugin.plugin_module
        elif pluginname == "self":
            # these are both from python's stdlib, the types aren't compatible
            # but it does seem to work
            srcfile = getfile(currentframe())  # type: ignore [arg-type]
        else:
            mesg = "Plugin <%s> not found" % pluginname
            raise PluginManagerException(mesg)
        version = repository.get_version(srcfile)
        __version_cache[pluginname] = version
    return version


def get_version(pluginname: str) -> tuple[int, int, int]:
    """Returns the internal version of a tool (index in datatable)

    If the version is not indexed it will be created.

    Parameters
    ----------
    pluginname
        Name of the plugin

    Returns
    -------
    tuple[int, int, int]
        Tuple of (major, minor, patch) versions
    """
    tool_name = pluginname.lower()
    p = get_plugin_instance(pluginname)
    version = repr(p.__version__)
    (repos_tool, version_tool) = get_plugin_version(pluginname)
    version_id = (
        User()
        .getUserDB()
        .getVersionId(tool_name, version, "", version_api, repos_tool, version_tool)
    )
    if version_id is None:
        version_id = (
            User()
            .getUserDB()
            .newVersion(tool_name, version, "", version_api, repos_tool, version_tool)
        )
    return version_id


def dict2conf(
    toolname: str, conf_dict: dict[str, str], user: Optional[User] = None
) -> list[Configuration]:
    """Get a list of configuration model objects.

    Useful for getting similar results.

    Parameters
    ----------
    conf_dict
        dictionary with configuration to look up

    Returns
    -------
    list[Configuration]
        list of configuration objects
    """
    user = user or User()

    conf = []
    paramstring = []
    tool = get_plugin_instance(toolname, user)
    for key, value in conf_dict.items():
        o = Parameter.objects.filter(tool=toolname, parameter_name=key).order_by("-id")
        if len(o) == 0:
            string = "Parameter <%s> not found" % key
            raise ParameterNotFoundError(string)

        else:
            paramstring = ["%s=%s" % (key, str(value))]
            realvalue = tool.__parameters__.parse_arguments(
                paramstring, check_errors=False
            )[key]
            conf_object = Configuration()
            conf_object.parameter_id_id = o[0].id
            conf_object.value = json.dumps(realvalue)
            conf.append(conf_object)
    return conf


def plugin_env_iter(envvar: str) -> Iterator[tuple[str, ...]]:
    """Splits the elements of a plugin env string.

    Returns an interator over all elements in a plugin environment
    variable string

    Parameters
    ----------
    envvar
        string to split

    Returns
    -------
    Iterator[tuple[str, ...]]
        Iterator over the (path, package) elements found in the string.
        The type is variable length but this will only ever return
        a 2 element tuple when given a well formed string.
    """
    return map(
        lambda item: tuple([e.strip() for e in item.split(",")]),
        envvar.split(":"),
    )


def find_plugin_class(mod: ModuleType) -> type[PluginAbstract]:
    """Looks for a subclass of PluginAbstract within a module.

    Only returns the first one found if there are multiple.

    Parameters
    ----------
    mod
        module to search for plugin

    Returns
    -------
    type[PluginAbstract]
        The subclass of PluginAbstract found

    Raises
    ------
    PluginManagerException
        Raised when no subclass of PluginAbstract is found
    """
    for attr_name in dir(mod):
        attr = getattr(mod, attr_name)
        if inspect.isclass(attr) and attr.__base__ == PluginAbstract:
            return attr
    raise PluginManagerException()


# This only runs once after start. To load new plugins on the fly we have
# 2 possibilities:
# 1) Watch the tool directory
# 2) Use the plugin metaclass trigger (see `evaluation_system.api.plugin`
reload_plugins()



"""
Created on 11.03.2013

@author: Sebastian Illing / estani

This package encapsulate access to a solr instance (not for search but for administration)
We define two cores::

* files: all files  - id is file (full file path)
* latest: only those files from the latest dataset version - id is file_no_version (full file path *wothout* version information)

"""
from __future__ import annotations
import os
import shutil
import urllib
import urllib.request
import json
from datetime import datetime
from pathlib import Path
from typing import Dict, Iterator, Optional, Tuple

from evaluation_system.model.file import DRSFile
from evaluation_system.misc import config, logger as log
from evaluation_system.misc.exceptions import CommandError


class SolrCore:
    """Encapsulate access to a Solr instance"""

    def __init__(
        self,
        core=None,
        host=None,
        port=None,
        instance_dir=None,
        data_dir=None,
        get_status=True,
    ):
        """Create the connection pointing to the proper solr url and core.

        :param core: The name of the core referred (default: loaded from config file)
        :param host: the hostname of the Solr server (default: loaded from config file)
        :param port: The port number of the Solr Server (default: loaded from config file)
        :param instance_dir: the core instance directory (if empty but the core exists it will get downloaded from Solr)
        :param data_dir: the directory where the data is being kept (if empty but the core exists it will
        get downloaded from Solr)"""

        self.host = host or config.get(config.SOLR_HOST)
        self.port = port or config.get(config.SOLR_PORT)
        self.core = core or config.get(config.SOLR_CORE)
        self.solr_url = f"http://{self.host}:{self.port}/solr/"
        self.core_url = self.solr_url + self.core + "/"
        self.instance_dir = instance_dir
        self.data_dir = data_dir

        if get_status:
            st = self.status()
        else:
            st = {}
        if self.instance_dir is None and "instanceDir" in st:
            self.instance_dir = st["instanceDir"]
        if self.data_dir is None and "dataDir" in st:
            self.data_dir = st["dataDir"]
        else:
            self.data_dir = "data"

        # Other Defaults
        import socket

        socket.setdefaulttimeout(20)

    def __str__(self):
        return "<SolrCore %s>" % self.core_url

    def post(self, list_of_dicts, auto_list=True, commit=True):
        """Sends some json to Solr for ingestion.

        :param list_of_dicts: either a json or more normally a list of json instances that will be sent to Solr for ingestion
        :param auto_list: avoid packing list_of dicts in a directory if it's not one
        :param commit: send also a Solr commit so that changes can be seen immediately."""
        if auto_list and not isinstance(list_of_dicts, list):
            list_of_dicts = [list_of_dicts]
        endpoint = "update/json?"
        if commit:
            endpoint += "commit=true"
        query = self.core_url + endpoint
        log.debug(query)
        post_data = json.dumps(list_of_dicts).encode("ascii")
        req = urllib.request.Request(query, post_data)
        req.add_header("Content-type", "application/json")

        return urllib.request.urlopen(req).read()

    def get_json(self, endpoint, use_core=True, check_response=True):
        """Return some json from server. Is the raw access to Solr.

        :param endpoint: The endpoint, path missing after the core url and all parameters encoded in it (e.g. 'select?q=*')
        :param use_core: if the core info is used for generating the endpoint. (if False, then == self.core + '/' + endpoint)
        :param check_response: If the response should be checked for errors. If True, raise an exception if something is
         wrong (default: True)"""
        if "?" in endpoint:
            endpoint += "&wt=json"
        else:
            endpoint += "?wt=json"

        if use_core:
            query = self.core_url + endpoint
        else:
            query = self.solr_url + endpoint
        log.debug(query)

        try:
            req = urllib.request.Request(query)
            response = json.loads(urllib.request.urlopen(req).read())
        except urllib.error.HTTPError:
            raise CommandError("Bad request option")
        if response["responseHeader"]["status"] != 0:
            raise CommandError(
                "Error while accessing Core %s. Response: %s" % (self.core, response)
            )

        return response

    def get_solr_fields(self):
        """Return information about the Solr fields. This is dynamically generated and because of
        dynamicFiled entries in the Schema, this information cannot be inferred from anywhere else."""
        answer = self.get_json("admin/luke")["fields"]
        # TODO: Solr has a language facet. Until we know why, delete it
        if isinstance(answer, dict):
            try:
                del answer["language"]
            except KeyError:
                pass
        else:
            try:
                answer.remove("language")
            except ValueError:
                pass
        return answer

    def create(
        self,
        instance_dir=None,
        data_dir=None,
        config="solrconfig.xml",
        schema="schema.xml",
        check_if_exist=True,
    ):
        """Creates (actually "register") this core. The Core configuration and directories must
        be generated beforehand (not the data one). You may clone an existing one or start from scratch.

        :param instance_dir: main directory for this core
        :param data_dir: Data directory for this core (if left unset, a local "data" directory in instance_dir will be used)
        :param config: The configuration file (expected in instance_dir/conf)
        :param schema: The schema file (expected in instance_dir/conf)
        :param check_if_exist: check for the existence of the instance directorie"""
        # check basic configuration (it must exists!)
        if instance_dir is None and self.instance_dir is None:
            raise ValueError("No Instance directory defined!")
        elif instance_dir is not None:
            self.instance_dir = instance_dir
        if not os.path.isdir(self.instance_dir) and check_if_exist:
            raise FileNotFoundError(
                "Expected Solr Core configuration not found in %s" % self.instance_dir
            )

        if data_dir is not None:
            self.data_dir = data_dir

        return self.get_json(
            "admin/cores?action=CREATE&name=%s" % self.core
            + "&instanceDir=%s" % self.instance_dir
            + "&config=%s" % config
            + "&schema=%s" % schema
            + "&dataDir=%s" % self.data_dir,
            use_core=False,
        )

    def reload(self):
        """Reload the core. Useful after schema changes.
        Be aware that you might need to re-ingest everything if there were changes to the indexing part of the schema."""
        return self.get_json(
            "admin/cores?action=RELOAD&core=" + self.core, use_core=False
        )

    def unload(self):
        """Unload the core."""
        return self.get_json(
            "admin/cores?action=UNLOAD&core=" + self.core, use_core=False
        )

    def swap(self, other_core):
        """Will swap this core with the given one (that means rename their references)

        :param other_core: the name of the other core that this will be swapped with."""
        return self.get_json(
            "admin/cores?action=SWAP&core=%s&other=%s" % (self.core, other_core),
            use_core=False,
        )

    def status(self, general=False):
        """Return status information about this core or the whole Solr server.

        :param general: If True return all information as provided by the server, otherwise just the status info
        from this core."""
        url_str = "admin/cores?action=STATUS"
        if not general:
            url_str += "&core=" + self.core
        response = self.get_json(url_str, use_core=False)
        if general:
            return response
        else:
            return response["status"][self.core]

    def clone(self, new_instance_dir, data_dir="data", copy_data=False):
        """Copies a core somewhere else.
        :param new_instance_dir: the new location for the clone.
        :param data_dir: the location of the data directory for this new clone.
        :param copy_data: If the data should also be copied (Warning, this is done on the-fly so be sure to unload the core
        first) or assure otherwise there's no chance of getting corrupted data (I don't know any other way besides
        unloading the original code))"""
        try:
            os.makedirs(new_instance_dir)
        except:
            pass
        shutil.copytree(
            os.path.join(self.instance_dir, "conf"),
            os.path.join(new_instance_dir, "conf"),
        )
        if copy_data:
            shutil.copytree(
                os.path.join(self.instance_dir, self.data_dir),
                os.path.join(new_instance_dir, data_dir),
            )

    def delete(self, query):
        """Issue a delete command, there's no default query for this to avoid unintentional deletion."""
        self.post(dict(delete=dict(query=query)), auto_list=False)

    @staticmethod
    def _get_metadata_from_path(
        in_dir: Path,
        abort_on_errors: bool,
        allowed_suffixes: Tuple[str, ...],
        drs_type: Optional[str] = None,
    ) -> Iterator[Tuple[DRSFile, Dict[str, str]]]:
        if in_dir.is_file():
            iterator = [in_dir]
        else:
            iterator = dir_iter(in_dir)
        for file in iterator:
            if file.suffix not in allowed_suffixes:
                continue
            timestamp = file.stat().st_mtime
            try:
                drs_file = DRSFile.from_path(file, activity=drs_type)
            except (ValueError, FileNotFoundError) as e:
                if abort_on_errors:
                    raise e
                log.error(e.__str__())
                continue
            metadata = SolrCore.to_solr_dict(drs_file)
            metadata["timestamp"] = timestamp
            metadata["creation_time"] = timestamp_to_solr_date(timestamp)
            yield drs_file, metadata

    def _del_file_pattern(self, file_pattern: Path, prefix: str = "file") -> None:
        """Delete all entries of the core."""
        file_pattern = Path(file_pattern).expanduser().absolute()
        # TODO: Better way to determine if we have a regex on board
        if file_pattern.is_dir():
            file_pattern /= "*"
        self.delete(f"{prefix}:\\{file_pattern}")

    @staticmethod
    def delete_entries(
        file_pattern: Path,
        host: Optional[str] = None,
        port: Optional[int] = None,
        prefix: str = "file",
    ) -> None:
        """Delete all corresponding entries the the solr server.

        Parameters:
        ----------
        file_pattern:
            The input directory which contains the files to be delted from the
            solr server
        host:
            The server hostname of the apache solr server.
        port:
            The host port number the apache solr server is listinig to.
        prefix:
            The prefix representing the data sotre, currently only posix file
            types are supported (file)
        """
        core_latest = SolrCore(core="latest", host=host, port=port)
        core_all_files = SolrCore(core=None, host=host, port=port)
        core_all_files._del_file_pattern(file_pattern)
        core_latest._del_file_pattern(file_pattern)

    @staticmethod
    def load_fs(
        input_dir: Path,
        drs_type: Optional[str] = None,
        chunk_size: int = 10000,
        suffix: Tuple[str, ...] = (".nc", ".grb", ".zarr", ".grib", ".nc4"),
        core: Optional[str] = None,
        core_latest: Optional[SolrCore] = None,
        core_all_files: Optional[SolrCore] = None,
        abort_on_errors: bool = False,
        host: Optional[str] = None,
        port: Optional[int] = None,
    ) -> None:
        """Load information of files on posix file system into Solr.

        This method loads the information from a file and decides if it should be added
        to just the common file core, holding the index of all files,
        or also to the *latest* core, holding information
        about the latest version of all files
        remember that in CMIP5 not all files are version, just the datasets).

        Parameters:
        -----------
        input_dir:
            Directory or input file that is crawled
        chunk_size:
            Number of entries that will be written to the Solr main core
             (the latest core will be flushed at the same time and is
             guaranteed to have at most as many as the other.)
        abort_on_errors:
            If dumping should get aborted as soon as an error is found,
            i.e. a file that can't be ingested. Most of the times there are many
            files being found in the dump file that are no data at all
        drs_type:
            Pre-define the data type to search for. If None (default) try
            guessing the type
        suffix:
            The file types that are taken into account when searching for data
        host:
            The server hostname of the apache solr server.
        port:
            The host port number the apache solr server is listinig to."""
        core_latest = core_latest or SolrCore(core="latest", host=host, port=port)
        core_all_files = core_all_files or SolrCore(core=core, host=host, port=port)
        core_latest._del_file_pattern(input_dir)
        core_all_files._del_file_pattern(input_dir)
        chunk, chunk_latest = [], []
        chunk_count = 0
        chunk_latest_new: Dict[str, Dict[str, str]] = {}
        latest_versions: Dict[str, str] = {}
        for (drs_file, metadata) in SolrCore._get_metadata_from_path(
            input_dir, abort_on_errors, suffix, drs_type=drs_type
        ):
            chunk.append(metadata)
            if drs_file.versioned:
                # TODO: We need a proper data set versioning.
                version = latest_versions.get(
                    drs_file.to_dataset(versioned=False), "-1"
                )
                idx = drs_file.to_dataset(versioned=False)
                if (drs_file.version or "0") > version:
                    # unknown or new version, update
                    version = drs_file.version or "0"
                    latest_versions[idx] = version
                    chunk_latest_new[idx] = metadata
                if (drs_file.version or "0") >= version:
                    chunk_latest.append(metadata)
            else:
                # if not version always add to latest
                chunk_latest_new[drs_file.to_dataset(versioned=False)] = metadata
                chunk_latest.append(metadata)
            if len(chunk) >= chunk_size:
                log.info(
                    "Sending entries %s-%s"
                    % (chunk_count * chunk_size, (chunk_count + 1) * chunk_size)
                )
                core_all_files.post(chunk)
                chunk = []
                chunk_count += 1
                if chunk_latest:
                    core_latest.post(chunk_latest)
                    chunk_latest, chunk_latest_new = [], {}
        # flush
        if len(chunk) > 0:
            log.info("Sending last %s entries" % (len(chunk)))
            core_all_files.post(chunk)
            if chunk_latest:
                core_latest.post(chunk_latest)

    @staticmethod
    def to_solr_dict(drs_file):
        """Extracts from a DRSFile the information that will be stored in Solr"""
        metadata = drs_file.dict["parts"].copy()
        metadata["file"] = drs_file.to_path()
        if "version" in metadata:
            metadata["file_no_version"] = metadata["file"].replace(
                "/%s/" % metadata["version"], "/"
            )
        else:
            metadata["file_no_version"] = metadata["file"]
        metadata["dataset"] = drs_file.drs_structure
        return metadata


def timestamp_to_solr_date(timestamp):
    """Transform a timestamp (float) into a string parsable by Solr"""
    return datetime.fromtimestamp(timestamp).strftime("%Y-%m-%dT%H:%M:%SZ")


def dir_iter(start_dir, abort_on_error=True, followlinks=True):
    for base_dir, dirs, files in os.walk(start_dir, followlinks=followlinks):
        # make sure we walk them in the proper order (latest version first)
        dirs.sort(reverse=True)
        files.sort(reverse=True)  # just for consistency
        for f in files:
            yield Path(base_dir) / f



"""Collection of admin commands that perform checks."""
from __future__ import annotations
import argparse
import os
from typing import Any
from ..utils import BaseParser, is_admin, subparser_func_type

from evaluation_system.misc import logger
from evaluation_system.misc.exceptions import CommandError
from evaluation_system.model.history.models import History
from evaluation_system.model.plugins.models import ToolPullRequest
from evaluation_system.api import plugin_manager as pm


__all__ = ["check4broken_runs", "check4pull_request"]

CLI = "CheckCli"


def check4pull_request() -> None:
    """Check for pending pull requests."""

    is_admin(raise_error=True)
    pull_requests = ToolPullRequest.objects.filter(status="waiting")
    tools = pm.get_plugins()
    for request in pull_requests:
        print(f"Processing pull request for {request.tool} by {request.user}")
        request.status = "processing"
        request.save()
        tool_name = request.tool.lower()
        if tool_name not in tools.keys():
            request.status = "failed"
            request.save()
            logger.error(f"Plugin {request.tool} does not exist")
            return
        # get repo path
        path = "/".join(tools[tool_name].plugin_module.split("/")[:-1])
        exit_code = os.system(
            "cd %s; git pull; git checkout -b version_%s %s"
            % (path, request.tagged_version, request.tagged_version)
        )
        if exit_code > 1:
            # Probably branch exists already
            # Try to checkout old branch
            exit_code = os.system(
                "cd %s; git checkout version_%s " % (path, request.tagged_version)
            )
            if exit_code > 1:
                raise CommandError("Something went wrong, please contact the admins")
        request.status = "success"
        request.save()


def check4broken_runs() -> None:
    """Check for broken runs in SLURM"""

    is_admin(raise_error=True)
    running_jobs = History.objects.filter(status=History.processStatus.running).exclude(
        slurm_output=0
    )
    for job in running_jobs:
        slurm_status = job.get_slurm_status()
        if (
            "cancelled" in slurm_status.lower()
            or "timeout" in slurm_status.lower()
            or "fail" in slurm_status.lower()
        ):
            print(slurm_status, job.tool)
            print(f"Setting job {job.id} to broken")
            job.status = job.processStatus.broken
            job.save()
        elif "completed" in slurm_status.lower():
            job.status = job.processStatus.finished
            job.save()


class PullRequest(BaseParser):
    """Command line interface to check for incoming PR's"""

    desc = "Check for incoming pull requests."

    def __init__(self, parser: argparse.ArgumentParser) -> None:
        """Construct the sub arg. parser."""

        parser.add_argument(
            "--deamon", help="Spawn in daemon mode", action="store_true", default=False
        )
        self.parser = parser
        parser.add_argument(
            "--debug",
            "--verbose",
            help="Use verbose output.",
            action="store_true",
            default=False,
        )
        self.logger.setLevel(20)  # Set log level to info
        self.parser.set_defaults(apply_func=self.run_cmd)

    @staticmethod
    def run_cmd(args: argparse.Namespace, **kwargs: Any) -> None:
        """Apply the check4broken_runs method"""

        check4pull_request()


class BrokenRun(BaseParser):
    """Command line interface to check for broken runs in batchmode."""

    desc = "Check for broken runs and report them."

    def __init__(self, parser: argparse.ArgumentParser) -> None:
        """Construct the sub arg. parser."""

        self.parser = parser
        parser.add_argument(
            "--debug",
            "--verbose",
            help="Use verbose output.",
            action="store_true",
            default=False,
        )
        self.logger.setLevel(20)  # Set log level to info
        self.parser.set_defaults(apply_func=self.run_cmd)

    @staticmethod
    def run_cmd(args: argparse.Namespace, **kwargs: Any) -> None:
        """Apply the check4broken_runs method"""

        check4broken_runs()


class CheckCli(BaseParser):
    """Interface defining parsers to perform checks."""

    desc = "Perform various checks."

    def __init__(self, parser: argparse.ArgumentParser) -> None:
        """Construct the sub arg. parser."""

        sub_commands: dict[str, subparser_func_type] = {
            "broken-runs": self.parse_broken_runs,
            "pull-request": self.parse_pull_request,
        }
        super().__init__(sub_commands, parser)
        # This parser doesn't do anything without a sub-commands
        # hence the default function should just print the usage
        self.parser.set_defaults(apply_func=self._usage)

    @staticmethod
    def parse_pull_request(
        help: str, subparsers: argparse._SubParsersAction
    ) -> PullRequest:
        sub_parser = subparsers.add_parser(
            "pull-request",
            description=help,
            help=help,
            formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        )
        return PullRequest(sub_parser)

    @staticmethod
    def parse_broken_runs(
        help: str, subparsers: argparse._SubParsersAction
    ) -> BrokenRun:
        sub_parser = subparsers.add_parser(
            "broken-runs",
            description=help,
            help=help,
            formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        )
        return BrokenRun(sub_parser)
